from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
import os
import yaml
from pathlib import Path
import glob
import dtiplayground.config as config
from dtiplayground.ui.protocoltab import ProtocolTab
from dtiplayground.ui.executiontab import ExecutionTab
from dtiplayground.ui.protocol import Protocol
from dtiplayground.ui.quickview import QuickView

class Widgets(QWidget):
    
  def __init__(self, parent, protocol_template, args):
    super(QWidget, self).__init__(parent)
    self.protocol_template = protocol_template
    self.unsaved_changes = True

    # User preferences
    user_directory = os.path.expanduser("~/.niral-dti")
    if not os.path.exists(user_directory + "/user_preferences.yml"):
      self.CreateUserPreferencesFile()
    preferences_yml = yaml.safe_load(open(user_directory + "/user_preferences.yml", 'r'))
    
    # Load yml file used to initialize modules
    def_protocol_path = Path(__file__).parent.joinpath('data/init_protocol.yml')
    protocol_yml = yaml.safe_load(open(def_protocol_path.__str__(),'r'))
    
    # instanciation of classes
    
    self.protocol_tab = ProtocolTab(protocol_template, protocol_yml, preferences_yml) #elements related to the protocol tab
    self.execution_tab = ExecutionTab(protocol_template, args) #elements related to the execution tab
    init_module_id = self.protocol_tab.selector.modules_list_widget.count()
    self.protocol = Protocol(protocol_template, protocol_yml, preferences_yml, init_module_id) #functions related to the manipulations of the protocol

    # Initialize tabs
    self.tabs = QTabWidget()    
    self.tabs.addTab(self.protocol_tab.tab, "Protocol")
    self.tabs.addTab(self.execution_tab.tab, "Execution")       

    # Add tabs to widget
    layout = QVBoxLayout(self)
    layout.addWidget(self.tabs)
    self.setLayout(layout)

    # Create red and green icons : could be a function
    pixmap = QPixmap(100, 100)
    pixmap.fill(QColor("red"))
    self.red_icon = QIcon(pixmap)
    pixmap.fill(QColor("green"))
    self.green_icon = QIcon(pixmap)

    # Communication between QuickView and Dmriprep
    QuickView.signal_quickview.execute_exclude_gradients.connect(self.ExecuteExcludeGradientsFromQuickView)
    # Communicatin between modules_details classes and Protocol class
    self.protocol_tab.details.baselineaverage.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.brainmask.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.dtiestimate.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.eddymotion.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.exclude.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.interlacecheck.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.slicecheck.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.susceptibility.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.susceptibility.communicate.call_update_user_preferences_file.connect(self.UpdateUserPreferencesFile)
    self.protocol_tab.details.utilheader.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.utilmerge.communicate.send_params.connect(self.UpdateModuleParams)
    self.protocol_tab.details.qcreport.communicate.send_params.connect(self.UpdateModuleParams)
    # Communication from module selector subtab
    self.protocol_tab.selector.communicate.module_added_to_protocol_drop.connect(self.protocol.ModuleAddedToProtocolDrop)
    self.protocol_tab.selector.communicate.call_execute_exclude_gradients_popup.connect(self.protocol_tab.exclude_popup.exec_)
    self.protocol_tab.selector.communicate.call_execute_merge_images_popup.connect(self.protocol_tab.merge_popup.exec_)
    # Communication from module details subtab
    self.protocol_tab.details.communicate.set_update_module_details.connect(self.SetUpdateModuleDetails)
    self.protocol_tab.details.communicate.call_OrderModulesInProtocolList.connect(self.GetDicProtocolKeys)
    self.protocol_tab.details.communicate.get_selected_item_data.connect(self.GetDicProtocolSelectedItem)
    # Communication from general subtab
    self.protocol_tab.general.communicate.call_UnsavedModif.connect(self.UnsavedModif)    
    # Communication from execution tab
    self.execution_tab.communicate.call_DeleteTempDirectory.connect(self.DeleteTempDirectory)
    self.execution_tab.communicate.check_manual_exclude.connect(self.CheckManualExcludeInProtocol)
    self.execution_tab.communicate.call_write_protocol_yml.connect(self.protocol.WriteProtocolYML)
    self.execution_tab.communicate.check_unsaved_changes.connect(self.CheckUnsavedChangesBeforeCompute)
    self.execution_tab.communicate.call_save_protocol.connect(self.protocol_tab.SaveProtocol)
    # Communication from protocol tab
    self.protocol_tab.communicate.call_module_added_to_protocol.connect(self.protocol.ModuleAddedToProtocol)
    self.protocol_tab.communicate.call_unsaved_modif.connect(self.UnsavedModif)
    self.protocol_tab.communicate.call_order_modules_in_protocol_list.connect(self.GetDicProtocolKeys)
    self.protocol_tab.communicate.set_loading_protocol.connect(self.protocol.SetLoadingProtocol)
    self.protocol_tab.communicate.set_module_id.connect(self.protocol.SetModuleId)
    self.protocol_tab.communicate.get_module_id.connect(self.GetModuleId)
    self.protocol_tab.communicate.call_set_default_protocol.connect(self.protocol.SettingDefaultProtocol)
    self.protocol_tab.communicate.call_read_protocol.connect(self.protocol.ReadProtocol)
    self.protocol_tab.communicate.call_remove_module_from_dic_protocol.connect(self.protocol.RemoveModuleFromDicProtocol)
    self.protocol_tab.communicate.enable_eddymotion_param.connect(self.EnableEddymotionParam)
    self.protocol_tab.communicate.set_dic_protocol.connect(self.protocol.SetDicProtocol)
    self.protocol_tab.communicate.call_no_unsaved_modif.connect(self.NoUnsavedModif)
    self.protocol_tab.communicate.call_write_protocol_yml.connect(self.protocol.WriteProtocolYML)
    self.protocol_tab.communicate.module_added_to_protocol_loading.connect(self.protocol.ModuleAddedToProtocolLoading)
    self.protocol_tab.communicate.call_update_user_preferences_file.connect(self.UpdateUserPreferencesFile)
    # Communication from protocol class
    self.protocol.communicate.call_clear_protocol.connect(self.protocol_tab.ClearProtocol)
    self.protocol.communicate.set_general_param_from_read_protocol.connect(self.SetGeneralParam)
    self.protocol.communicate.add_module_to_protocol_widget.connect(self.protocol_tab.AddModuleToProtocolWidget)
    self.protocol.communicate.get_protocol_io.connect(self.GetProtocolIO)
    self.protocol.communicate.call_unsaved_modif.connect(self.UnsavedModif)
    self.protocol.communicate.get_list_widget_modules.connect(self.protocol_tab.GetListWidgetModules)
    self.protocol.communicate.get_selected_modules.connect(self.protocol_tab.selector.GetListWidgetSelectedModules)
    self.protocol.communicate.set_module_data.connect(self.protocol_tab.SetModuleData)
    self.protocol.communicate.call_execute_exclude_gradients_popup.connect(self.protocol_tab.exclude_popup.exec_)
    self.protocol.communicate.call_execute_merge_images_popup.connect(self.protocol_tab.merge_popup.exec_)
    self.protocol.communicate.enable_modules_list_widget.connect(self.protocol_tab.selector.EnableModulesListWidget)
    self.protocol.communicate.enable_exclude_gradients_module.connect(self.protocol_tab.selector.EnableExcludeGradientsModule)
    self.protocol.communicate.enable_merge_images_module.connect(self.protocol_tab.selector.EnableMergeImagesModule)
    self.protocol.communicate.update_selector_data.connect(self.protocol_tab.selector.UpdateSelectorData)
    self.protocol.communicate.enable_eddymotion_param.connect(self.EnableEddymotionParam)


    self.tabs.tabBarClicked.connect(self.MultiInput)
    
    # Command Line arguments
    if args.protocol:
      #arg_protocol = yaml.safe_load(open(args.protocol,'r'))
      self.protocol.ReadProtocol(args.protocol)
    if args.output_directory:
      self.execution_tab.output_line.setText(args.output_directory)
    if args.image:
      if len(args.image) == 1:
        self.execution_tab.input_filename = args.image[0]
        if args.quickview:
          self.protocol.dic_protocol[0] = self.protocol.exclude_default_params
          self.execution_tab.multi_input = 0
          self.execution_tab.Compute([True, True]) #launch QuickView window
      else:
        self.execution_tab.input_filename1 = args.image[0]
        self.execution_tab.input_filename2 = args.image[1]


    self.unsaved_changes = False    

  def UnsavedModif(self):
    if self.unsaved_changes == False:
      self.unsaved_changes = True
      self.tabs.setTabIcon(0, self.red_icon)
      self.tabs.setTabToolTip(0, "Unsaved modifications have been made to protocol")

  def NoUnsavedModif(self):
    self.unsaved_changes = False
    self.tabs.setTabIcon(0, self.green_icon)
    self.tabs.setTabToolTip(0, "Protocol saved")

  def CreateUserPreferencesFile(self):
    user_directory = os.path.expanduser("~/.niral-dti")
    default = {"fslConfigurationFilePath": user_directory + "/dmriprep-" + config.INFO["dmriprep"]["version"] + "/parameters/fsl/fsl_regb02b0.cnf",
      "showExcludeGradientsPopup": True, 
      "showMergeImagesPopup": True}
    with open(user_directory + "/user_preferences.yml", 'w')as filename:
      yaml.dump(default, filename)

  def UpdateUserPreferencesFile(self, preferences):
    user_directory = os.path.expanduser("~/.niral-dti")
    with open(user_directory + "/user_preferences.yml", 'w')as filename:
      yaml.dump(preferences, filename)
    
  def DeleteTempDirectory(self):
    if os.path.exists("temp_dmriprep_ui/protocol.yml"):
      os.remove("temp_dmriprep_ui/protocol.yml")
    if os.path.exists("temp_dmriprep_ui"):
      os.rmdir("temp_dmriprep_ui")

  def ExecuteExcludeGradientsFromQuickView(self, list_gradients):
    key = list(self.protocol.dic_protocol.keys())[0]
    self.protocol.dic_protocol[key][2]["protocol"]['gradientsToExclude'] = list_gradients

    if not os.path.exists("temp_dmriprep_ui"):
      os.mkdir("temp_dmriprep_ui")
      print("Directory 'temp_dmriprep_ui' created")
    else:
      print("Directory 'temp_dmriprep_ui' already exists")
    self.protocol.WriteProtocolYML("temp_dmriprep_ui/protocol.yml")

    self.execution_tab.StartComputation("temp_dmriprep_ui/protocol.yml")

  def SetUpdateModuleDetails(self, bool_update_module_details):
    self.protocol.update_module_details = bool_update_module_details
  
  def GetDicProtocolSelectedItem(self, data):
    key = data[0]
    index = data[1]
    selected_module = self.protocol.dic_protocol[key]
    self.protocol_tab.details.communicate.SendDicProtocolSelectedItem([selected_module, index])

  def MultiInput(self):
    multi_input = 0
    for index in self.protocol.dic_protocol:
      for module_yml in self.protocol_tab.details.modules_yml_list:
        if self.protocol.dic_protocol[index][1] in module_yml["name"]:
          if "multi_input" in module_yml["process_attributes"]:
            multi_input = 1
            if self.protocol.dic_protocol[index][0] == "Merge Images":
              multi_input = 2

    self.execution_tab.communicate.NumberOfInputs(multi_input)

  def CheckManualExcludeInProtocol(self):
    computation_details = [False, False]
    if len(self.protocol.dic_protocol) > 0:
      key = list(self.protocol.dic_protocol.keys())[0]
      if self.protocol.dic_protocol[key][1] == "MANUAL_Exclude":
        computation_details[0] = True
        if self.protocol_tab.details.exclude.selectionMethod.currentText() == "QuickView":
          computation_details[1] = True
    self.execution_tab.communicate.CallCompute(computation_details)


  def CheckUnsavedChangesBeforeCompute(self):
    if self.unsaved_changes == True:
      self.execution_tab.unsaved_changes_popup.exec_()
    filename = self.protocol_tab.protocol_filename
    self.execution_tab.StartComputation(filename)

  def GetModuleId(self):
    identifier = self.protocol.module_id
    self.protocol_tab.DefaultProtocol(identifier)
  
  def SetModuleId(self, new_id):
    self.protocol.module_id = new_id

  def SetGeneralParam(self, protocol):
    self.execution_tab.output_line.setText(protocol["output_directory"])
    if protocol["no_output_image"] == False:
      self.protocol_tab.general.no_output_image_false.setChecked(True)
    else:
      self.protocol_tab.general.no_output_image_true.setChecked(True)
    image_type = ["output_format"]
    for ite in self.protocol_template["options"]["io"]["output_format"]["candidates"]:
      if ite["value"] == image_type:
        self.protocol_tab.general.output_format.setCurrentText(ite["caption"])
        self.protocol_tab.general.GetOutputImageTypeIt(ite["caption"])
    self.protocol_tab.general.b0.setValue(protocol["baseline_threshold"])

  def GetProtocolIO(self, data):
    data[0]["io"] = {
        'baseline_threshold': self.protocol_tab.general.b0.value(),
        'no_output_image': self.protocol_tab.general.no_output_image_true.isChecked(),
        'output_directory': '',
        'output_format': self.protocol_template["options"]["io"]["output_format"]["candidates"][self.protocol_tab.general.outputimagetype_it]["value"]
      }
    self.protocol.Writing(data)

  def GetIndexAndKey(self):
    index = self.protocol_tab.protocol_list_widget.currentRow() #in protocol_list_widget
    key = self.protocol_tab.protocol_list_widget.item(index).data(QtCore.Qt.UserRole) #module_id, in dic_protocol (dictionary)
    return index, key

  def UpdateModuleParams(self, params):
    self.UnsavedModif()
    index, key = self.GetIndexAndKey()
    if index >= 0 and self.protocol.update_module_details == True:
      self.protocol.dic_protocol[key] = params

  def GetDicProtocolKeys(self):
    list_new_keys = []
    for module_index in range(len(self.protocol.dic_protocol)): 
      list_new_keys.append(self.protocol_tab.protocol_list_widget.item(module_index).data(QtCore.Qt.UserRole))
    self.protocol.OrderModulesInProtocolList(list_new_keys)

  def EnableEddymotionParam(self, value):
    self.protocol_tab.details.eddymotion.estimateMoveBySusceptibility_true.setEnabled(value)
    self.protocol_tab.details.eddymotion.estimateMoveBySusceptibility_false.setEnabled(value)


from PyQt5.QtWidgets import * 
import os
import glob
import yaml
import dtiplayground.config as config

class SoftwarePaths(QMainWindow):
    def __init__(self):
        QMainWindow.__init__(self)

        #self.setFixedWidth(800)
        self.setWindowTitle("Softwares")
        widget = QWidget()
        self.setCentralWidget(widget)

        layout = QGridLayout()
        widget.setLayout(layout)

        self.software_paths_filepath = os.path.expanduser("~/.niral-dti/dmriprep-" + config.INFO["dmriprep"]["version"] + "/software_paths.yml")
        self.software_paths_yml = yaml.safe_load(open(self.software_paths_filepath, 'r'))

        list_softwares = list(self.software_paths_yml["softwares"].keys())
        self.softwares_dict = {}

        for i in range(len(list_softwares)):
            layout.addWidget(QLabel(list_softwares[i]), i, 0)
            path = QLineEdit()
            path.setText(self.software_paths_yml["softwares"][list_softwares[i]]["path"])
            layout.addWidget(path, i, 1, 1, 5)
            self.softwares_dict[list_softwares[i]] = path

        button_cancel = QPushButton("Cancel")
        layout.addWidget(button_cancel, i+1, 5)
        button_cancel.clicked.connect(lambda: self.close())

        button = QPushButton("Save")
        layout.addWidget(button, i+1, 4)
        button.clicked.connect(self.UpdatePaths)

    def UpdatePaths(self):
        for software in self.softwares_dict.items():
            self.software_paths_yml["softwares"][software[0]] = {"path": software[1].text()}
        with open(self.software_paths_filepath, 'w')as filename:
            yaml.dump(self.software_paths_yml, filename)

        self.close()






import dtiplayground.dmri.preprocessing as prep

import yaml,os
from pathlib import Path 

logger=prep.logger.write

### utilities

def find_fsl(lookup_dirs=[]):
    fsldir=os.environ.get('FSLDIR')
    candidates=[]
    if fsldir is not None:
        fsldir=Path(fsldir)
        candidates.append(fsldir)
    else:
        candidates=[]
        for d in lookup_dirs:
            candidates+=list(Path(d).glob("**/etc/fslversion"))
        
        if len(candidates)==0 : 
            logger("FSL6 NOT found",prep.Color.WARNING)
            return None,None 
        candidates=list(map(lambda x: x.parent.parent, candidates))

    fsl_version=None
    for d in candidates:
        fsldir=d
        versionfile=fsldir.joinpath('etc/fslversion')
        if versionfile.exists():
            with open(versionfile,'r') as f:
                fsl_version=f.readlines()[0]
            bigversion=int(fsl_version.split('.')[0])
            if bigversion>=6:
                return str(fsldir), fsl_version

    return str(fsldir),fsl_version


from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class DTIReg(ExternalToolWrapper):
    def __init__(self,binary_path):
        super().__init__(binary_path)

    def execute_with_args(self,args,**kwargs):
        self.setArguments(args)
        return self.execute(args)

    def compute_global_deformation_fields(self,
                                            fixed_volume,
                                            moving_volume,
                                            scalar_measurement,
                                            output_displacement_field,
                                            output_inverse_displacementField,
                                            output_volume,
                                            initial_affine,
                                            brains_transform,
                                            ants_outbase,
                                            program_paths:list,
                                            dti_reg_options:list,
                                            options=[]):

        arguments=[
            '--fixedVolume', fixed_volume,
            '--movingVolume', moving_volume,
            '--scalarMeasurement', scalar_measurement,
            '--outputDisplacementField', output_displacement_field,
            '--outputInverseDeformationFieldVolume', output_inverse_displacementField,
            '--outputVolume', output_volume,
            '--ProgramsPathsVector', ",".join(program_paths)
        ]
        m_DTIRegOptions=dti_reg_options 

        if m_DTIRegOptions[0]=="BRAINS":
          options+=["--method","useScalar-BRAINS"]
          if m_DTIRegOptions[1]=="GreedyDiffeo (SyN)":
            options+=["--BRAINSRegistrationType","GreedyDiffeo"]
          elif m_DTIRegOptions[1]=="SpatioTempDiffeo":
            options+=["--BRAINSRegistrationType","SpatioTempDiffeo"]
          else:
            options+=["--BRAINSRegistrationType",m_DTIRegOptions[1]]
          options.append(" --BRAINSnumberOfPyramidLevels " + m_DTIRegOptions[3] + "")
          options.append(" --BRAINSarrayOfPyramidLevelIterations " + m_DTIRegOptions[4] + "")
          if m_DTIRegOptions[2]=="Use computed affine transform":
            options.append(" --initialAffine " + initial_affine)
          else:
            options.append(" --BRAINSinitializeTransformMode " + m_DTIRegOptions[2] + "")
          #BRAINSTempTfm = FinalResampPath.joinpath("First_Resampling/" + case_id + "_" + scalar_measurement_type + "_AffReg.txt").__str__()
          options.append(" --outputTransform " + brains_transform)

        if m_DTIRegOptions[0]=="ANTS":
          options+=["--method","useScalar-ANTS"]
          if m_DTIRegOptions[1]=="GreedyDiffeo (SyN)":
            options+=["--ANTSRegistrationType","GreedyDiffeo"]
          elif m_DTIRegOptions[1]=="SpatioTempDiffeo (SyN)":
            options+=["--ANTSRegistrationType","SpatioTempDiffeo"]
          else:
            options+=["--ANTSRegistrationType",m_DTIRegOptions[1]]
          options+=["--ANTSTransformationStep",m_DTIRegOptions[2]]
          options+=["--ANTSIterations",m_DTIRegOptions[3]]
          if m_DTIRegOptions[4]=="Cross-Correlation (CC)" :
            options+=["--ANTSSimilarityMetric","CC"]
          elif m_DTIRegOptions[4]=="Mutual Information (MI)" :
            options+=["--ANTSSimilarityMetric","MI"]
          elif m_DTIRegOptions[4]=="Mean Square Difference (MSQ)":
            options+=["--ANTSSimilarityMetric","MSQ"]
          options+=["--ANTSSimilarityParameter",m_DTIRegOptions[5]]
          options+=["--ANTSGaussianSigma",m_DTIRegOptions[6]]
          if m_DTIRegOptions[7]=="1":
            options+=["--ANTSGaussianSmoothingOff"]

          options+=["--initialAffine",initial_affine]
          options+=["--ANTSUseHistogramMatching"]
          #ANTSTempFileBase = FinalResampPath.joinpath("First_Resampling/" + case_id + "_" + scalar_measurement_type + "_").__str__()
          options+=["--ANTSOutbase",ants_outbase]

        arguments+=options 
        self.setArguments(arguments)
        return self.execute(arguments)



from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from pathlib import Path
import yaml
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal as Signal

from dtiplayground.ui.modules_details.baselineaverage import BaselineAverage
from dtiplayground.ui.modules_details.brainmask import BrainMask
from dtiplayground.ui.modules_details.dtiestimate import DTIEstimate
from dtiplayground.ui.modules_details.eddymotion import EddyMotion
from dtiplayground.ui.modules_details.exclude import ExcludeGradients
from dtiplayground.ui.modules_details.interlacecheck import InterlaceCheck
from dtiplayground.ui.modules_details.slicecheck import SliceCheck
from dtiplayground.ui.modules_details.susceptibility import SusceptibilityCorrection
from dtiplayground.ui.modules_details.utilheader import UtilHeader
from dtiplayground.ui.modules_details.nomodule import NoModule
from dtiplayground.ui.modules_details.utilmerge import UtilMerge
from dtiplayground.ui.modules_details.qcreport import QCReport


class DetailsDisplayCommunicate(QObject):

  set_modules_details_tab = Signal()
  set_update_module_details = Signal(bool)
  call_OrderModulesInProtocolList = Signal()
  get_selected_item_data = Signal(list)
  get_dic_protocol_selected_item = Signal(list)

  def __init__(self):
    super(DetailsDisplayCommunicate, self).__init__()

  def ChangeTab(self):
    self.set_modules_details_tab.emit()

  def SetUpdateModuleDetailsBool(self, new_update_module_details):
    self.set_update_module_details.emit(new_update_module_details)

  def CallOrderModulesInProtocolList(self):
    self.call_OrderModulesInProtocolList.emit()

  def SendSelectedItemData(self, data):
    self.get_selected_item_data.emit(data)

  def SendDicProtocolSelectedItem(self, module):
    self.get_dic_protocol_selected_item.emit(module)

class ModuleDetails(QWidget):
  communicate = DetailsDisplayCommunicate()

  def __init__(self, protocol_template, preferences_yml):
    QWidget.__init__(self)
    self.protocol_template = protocol_template
    self.tab = QWidget()
    self.LoadModulesYMLFiles()
    self.ModuleDetails(protocol_template, preferences_yml)
    self.communicate.get_dic_protocol_selected_item.connect(self.DetailsDisplay)

     
  def ModuleDetails(self, protocol_template, preferences_yml):
    # instanciation of module classes
    self.baselineaverage = BaselineAverage(protocol_template, self.baselineaverage_yml)
    self.brainmask = BrainMask(protocol_template, self.brainmask_yml)
    self.dtiestimate = DTIEstimate(protocol_template, self.dtiestimate_yml)
    self.eddymotion = EddyMotion(protocol_template, self.eddymotion_yml)
    self.exclude = ExcludeGradients(protocol_template, self.exclude_yml)
    self.interlacecheck = InterlaceCheck(protocol_template, self.interlacecheck_yml)
    self.slicecheck = SliceCheck(protocol_template, self.slicecheck_yml)
    self.susceptibility = SusceptibilityCorrection(protocol_template, self.susceptibility_yml, preferences_yml)
    self.utilheader = UtilHeader(protocol_template, self.utilheader_yml)
    self.no_module = NoModule()
    self.utilmerge = UtilMerge(protocol_template, self.utilmerge_yml)
    self.qcreport = QCReport(protocol_template, self.qcreport_yml)

    # add modules to stack
    self.details_stack = QStackedWidget()
    self.details_stack.addWidget(self.no_module.stack)
    self.details_stack.addWidget(self.slicecheck.stack)
    self.details_stack.addWidget(self.interlacecheck.stack)    
    self.details_stack.addWidget(self.baselineaverage.stack)    
    self.details_stack.addWidget(self.susceptibility.stack)
    self.details_stack.addWidget(self.eddymotion.stack)
    self.details_stack.addWidget(self.brainmask.stack)
    self.details_stack.addWidget(self.dtiestimate.stack)
    self.details_stack.addWidget(self.exclude.stack)
    self.details_stack.addWidget(self.utilheader.stack)
    self.details_stack.addWidget(self.utilmerge.stack)
    self.details_stack.addWidget(self.qcreport.stack)

    # layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.details_stack)
    self.tab.setLayout(layout_v)

  def ModuleDetailsClicked(self, index_tab):
    if index_tab == 1:
      self.details_stack.setCurrentIndex(0)

  def LoadModulesYMLFiles(self):
    import dtiplayground.dmri.preprocessing.modules
    module_dir = Path(dtiplayground.dmri.preprocessing.modules.__file__).parent

    filepath = module_dir.joinpath("SLICE_Check/SLICE_Check.yml")
    self.slicecheck_yml = yaml.safe_load(open(filepath,'r'))

    filepath = module_dir.joinpath("INTERLACE_Check/INTERLACE_Check.yml")
    self.interlacecheck_yml = yaml.safe_load(open(filepath,'r'))

    filepath = module_dir.joinpath("BASELINE_Average/BASELINE_Average.yml")
    self.baselineaverage_yml = yaml.safe_load(open(filepath,'r'))

    filepath = module_dir.joinpath("SUSCEPTIBILITY_Correct/SUSCEPTIBILITY_Correct.yml")
    self.susceptibility_yml = yaml.safe_load(open(filepath,'r'))

    filepath = module_dir.joinpath("EDDYMOTION_Correct/EDDYMOTION_Correct.yml")
    self.eddymotion_yml = yaml.safe_load(open(filepath,'r'))

    filepath = module_dir.joinpath("BRAIN_Mask/BRAIN_Mask.yml")
    self.brainmask_yml = yaml.safe_load(open(filepath,'r'))

    filepath = module_dir.joinpath("DTI_Estimate/DTI_Estimate.yml")
    self.dtiestimate_yml = yaml.safe_load(open(filepath,'r'))

    filepath = module_dir.joinpath("MANUAL_Exclude/MANUAL_Exclude.yml")
    self.exclude_yml = yaml.safe_load(open(filepath,'r'))

    filepath = module_dir.joinpath("UTIL_Header/UTIL_Header.yml")
    self.utilheader_yml = yaml.safe_load(open(filepath, 'r'))

    filepath = module_dir.joinpath("UTIL_Merge/UTIL_Merge.yml")
    self.utilmerge_yml = yaml.safe_load(open(filepath, 'r'))

    filepath = module_dir.joinpath("QC_Report/QC_Report.yml")
    self.qcreport_yml = yaml.safe_load(open(filepath, 'r'))

    self.modules_yml_list = [self.slicecheck_yml, self.interlacecheck_yml, self.baselineaverage_yml,
      self.susceptibility_yml, self.eddymotion_yml, self.brainmask_yml, self.dtiestimate_yml, self.exclude_yml,
      self.utilheader_yml, self.utilmerge_yml, self.qcreport_yml]

  def DetailsDisplay(self, data):
    module = data[0]
    index = data[1]

    self.communicate.ChangeTab() #select details tab
    
    self.communicate.SetUpdateModuleDetailsBool(False)
    self.communicate.CallOrderModulesInProtocolList()

    if module[0] == "Slicewise Check":
      self.slicecheck.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.slicecheck.overwrite.setChecked(True)
      else:
        self.slicecheck.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.slicecheck.skip.setChecked(True)
      else:
        self.slicecheck.skip.setChecked(False)
      if module[2]["options"]["write_image"] == True:
        self.slicecheck.writeimage.setChecked(True)
      else:
        self.slicecheck.writeimage.setChecked(False)
      if module[2]["protocol"]["bSubregionalCheck"] == True:
        self.slicecheck.bSubregionalCheck_true.setChecked(True)
      else:
        self.slicecheck.bSubregionalCheck_false.setChecked(True)
      self.slicecheck.subregionalCheckRelaxationFactor.setValue(module[2]["protocol"]["subregionalCheckRelaxationFactor"])
      self.slicecheck.checkTimes.setValue(module[2]["protocol"]["checkTimes"])
      self.slicecheck.headSkipSlicePercentage.setValue(module[2]["protocol"]["headSkipSlicePercentage"])
      self.slicecheck.tailSkipSlicePercentage.setValue(module[2]["protocol"]["tailSkipSlicePercentage"])
      self.slicecheck.correlationDeviationThresholdbaseline.setValue(module[2]["protocol"]["correlationDeviationThresholdbaseline"])
      self.slicecheck.correlationDeviationThresholdgradient.setValue(module[2]["protocol"]["correlationDeviationThresholdgradient"])
      if module[2]["protocol"]["quadFit"] == True:
        self.slicecheck.quadFit_true.setChecked(True)
      else:
        self.slicecheck.quadFit_false.setChecked(True)
      self.details_stack.setCurrentIndex(1)

    if module[0] == "Interlace Correlation Check":
      self.interlacecheck.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.interlacecheck.overwrite.setChecked(True)
      else:
        self.interlacecheck.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.interlacecheck.skip.setChecked(True)
      else:
        self.interlacecheck.skip.setChecked(False)
      if module[2]["options"]["write_image"] == True:
        self.interlacecheck.writeimage.setChecked(True)
      else:
        self.interlacecheck.writeimage.setChecked(False)
      self.interlacecheck.correlationThresholdBaseline.setValue(module[2]["protocol"]["correlationThresholdBaseline"])
      self.interlacecheck.correlationThresholdGradient.setValue(module[2]["protocol"]["correlationThresholdGradient"])
      self.interlacecheck.correlationDeviationBaseline.setValue(module[2]["protocol"]["correlationDeviationBaseline"])
      self.interlacecheck.correlationDeviationGradient.setValue(module[2]["protocol"]["correlationDeviationGradient"])
      self.interlacecheck.translationThreshold.setValue(module[2]["protocol"]["translationThreshold"])
      self.interlacecheck.rotationThreshold.setValue(module[2]["protocol"]["rotationThreshold"])
      self.details_stack.setCurrentIndex(2)

    if module[0] == "Baseline Average":
      self.baselineaverage.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.baselineaverage.overwrite.setChecked(True)
      else:
        self.baselineaverage.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.baselineaverage.skip.setChecked(True)
      else:
        self.baselineaverage.skip.setChecked(False)
      if module[2]["options"]["write_image"] == True:
        self.baselineaverage.writeimage.setChecked(True)
      else:
        self.baselineaverage.writeimage.setChecked(False)
      method = module[2]["protocol"]["averageMethod"]
      for ite1 in self.baselineaverage_yml["protocol"]["averageMethod"]["candidates"]:
        if ite1["value"] == method:
          self.baselineaverage.averageMethod.setCurrentText(ite1["caption"])
          self.baselineaverage.GetAvgMethodIt(ite1["caption"])
      method = module[2]["protocol"]["averageInterpolationMethod"]
      for ite1 in self.baselineaverage_yml["protocol"]["averageInterpolationMethod"]["candidates"]:
        if ite1["value"] == method:
          self.baselineaverage.averageInterpolationMethod.setCurrentText(ite1["caption"])
          self.baselineaverage.GetAvgInterpolMethodIt(ite1["caption"])
      self.baselineaverage.stopThreshold.setValue(module[2]["protocol"]["stopThreshold"])
      self.baselineaverage.maxIterations.setValue(module[2]["protocol"]["maxIterations"])
      self.baselineaverage.outputDWIFileNameSuffix.setText(module[2]["protocol"]["outputDWIFileNameSuffix"])
      self.details_stack.setCurrentIndex(3)

    if module[0] == "Susceptibility correction":
      self.susceptibility.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.susceptibility.overwrite.setChecked(True)
      else:
        self.susceptibility.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.susceptibility.skip.setChecked(True)
      else:
        self.susceptibility.skip.setChecked(False)
      if module[2]["options"]["write_image"] == True:
        self.susceptibility.writeimage.setChecked(True)
      else:
        self.susceptibility.writeimage.setChecked(False)
      for ite1 in module[2]["protocol"]["phaseEncodingAxis"]:
        if ite1 == 0:
          self.susceptibility.phaseEncodingAxis_p0.setChecked(True)
        if ite1 == 1:
          self.susceptibility.phaseEncodingAxis_p1.setChecked(True)
        if ite1 == 2:
          self.susceptibility.phaseEncodingAxis_p2.setChecked(True)
      self.susceptibility.phaseEncodingValue.setValue(module[2]["protocol"]["phaseEncodingValue"])
      self.susceptibility.configurationFilePath.setText(module[2]["protocol"]["configurationFilePath"])
      self.details_stack.setCurrentIndex(4)

    if module[0] == "Eddy motion Correction":
      self.eddymotion.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.eddymotion.overwrite.setChecked(True)
      else:
        self.eddymotion.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.eddymotion.skip.setChecked(True)
      else:
        self.eddymotion.skip.setChecked(False)
      if module[2]["options"]["write_image"] == True:
        self.eddymotion.writeimage.setChecked(True)
      else:
        self.eddymotion.writeimage.setChecked(False)
      if module[2]["protocol"]["estimateMoveBySusceptibility"] == True:
        self.eddymotion.estimateMoveBySusceptibility_true.setChecked(True)
      else:
        self.eddymotion.estimateMoveBySusceptibility_false.setChecked(True)
      if module[2]["protocol"]["interpolateBadData"] == True:
        self.eddymotion.interpolateBadData_true.setChecked(True)
      else:
        self.eddymotion.interpolateBadData_false.setChecked(True)
      if module[2]["protocol"]["dataIsShelled"] == True:
        self.eddymotion.dataIsShelled_true.setChecked(True)
      else:
        self.eddymotion.dataIsShelled_false.setChecked(True)
      if module[2]["protocol"]["qcReport"] == True:
        self.eddymotion.qcReport_true.setChecked(True)
      else:
        self.eddymotion.qcReport_false.setChecked(True)
      self.details_stack.setCurrentIndex(5)

    if module[0] == "Brain Masking":
      self.brainmask.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.brainmask.overwrite.setChecked(True)
      else:
        self.brainmask.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.brainmask.skip.setChecked(True)
      else:
        self.brainmask.skip.setChecked(False)
      method = module[2]["protocol"]["method"]
      for ite1 in self.brainmask_yml["protocol"]["method"]["candidates"]:
        if ite1["value"] == method:
          self.brainmask.method.setCurrentText(ite1["caption"])
          self.brainmask.GetMethodIt(ite1["caption"])
      averagingmethod = module[2]["protocol"]["averagingMethod"]
      for ite1 in self.brainmask_yml["protocol"]["averagingMethod"]["candidates"]:
        if ite1["value"] == averagingmethod:
          self.brainmask.averagingmethod.setCurrentText(ite1["caption"])
          self.brainmask.GetAveragingMethodIt(ite1["caption"])
      if module[2]["protocol"]["modality"] == "t2":
        self.brainmask.modality_t2.setChecked(True)
      if module[2]["protocol"]["modality"] == "fa":
        self.brainmask.modality_fa.setChecked(True)
      self.details_stack.setCurrentIndex(6)

    if module[0] == "Estimate DTI":
      self.dtiestimate.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.dtiestimate.overwrite.setChecked(True)
      else:
        self.dtiestimate.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.dtiestimate.skip.setChecked(True)
      else:
        self.dtiestimate.skip.setChecked(False)
      if module[2]["options"]["write_image"] == True:
        self.dtiestimate.writeimage.setChecked(True)
      else:
        self.dtiestimate.writeimage.setChecked(False)
      method = module[2]["protocol"]["correctionMethod"]
      for ite1 in self.dtiestimate_yml["protocol"]["correctionMethod"]["candidates"]:
        if ite1["value"] == method:
          self.dtiestimate.correctionMethod.setCurrentText(ite1["caption"])
          self.dtiestimate.GetCorrectionMethodIt(ite1["caption"])
      method = module[2]["protocol"]["method"]
      for ite1 in self.dtiestimate_yml["protocol"]["method"]["candidates"]:
        if ite1["value"] == method:
          self.dtiestimate.method.setCurrentText(ite1["caption"])
          self.dtiestimate.GetMethodIt(ite1["caption"])
      method = module[2]["protocol"]["optimizationMethod"]
      for ite1 in self.dtiestimate_yml["protocol"]["optimizationMethod"]["candidates"]:
        if ite1["value"] == method:
          self.dtiestimate.optimizationMethod.setCurrentText(ite1["caption"])
          self.dtiestimate.GetOptimizationMethodIt(ite1["caption"])
      self.details_stack.setCurrentIndex(7)

    if module[0] == "Exclude Gradients":
      self.exclude.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.exclude.overwrite.setChecked(True)
      else:
        self.exclude.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.exclude.skip.setChecked(True)
      else:
        self.exclude.skip.setChecked(False)
      if module[2]["options"]["write_image"] == True:
        self.exclude.writeimage.setChecked(True)
      else:
        self.exclude.writeimage.setChecked(False)
      gradients_list = module[2]["protocol"]["gradientsToExclude"]
      gradients_string = ", ".join(str(gradient) for gradient in gradients_list)
      self.exclude.gradients2exclude.insertPlainText(gradients_string)
      self.details_stack.setCurrentIndex(8)

    if module[0] == "View Header":
      self.utilheader.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.utilheader.overwrite.setChecked(True)
      else:
        self.utilheader.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.utilheader.skip.setChecked(True)
      else:
        self.utilheader.skip.setChecked(False)
      if module[2]["options"]["write_image"] == True:
        self.utilheader.writeimage.setChecked(True)
      else:
        self.utilheader.writeimage.setChecked(False)
      self.utilheader.options.setText(module[2]["protocol"]["options"])
      self.details_stack.setCurrentIndex(9)

    if module[0] == "Merge Images":
      self.utilmerge.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.utilmerge.overwrite.setChecked(True)
      else:
        self.utilmerge.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.utilmerge.skip.setChecked(True)
      else:
        self.utilmerge.skip.setChecked(False)
      if module[2]["options"]["write_image"] == True:
        self.utilmerge.writeimage.setChecked(True)
      else:
        self.utilmerge.writeimage.setChecked(False)
      self.utilmerge.testparam.setText(module[2]["protocol"]["TestParam"])
      self.details_stack.setCurrentIndex(10)

    if module[0] == "QC Report":
      self.qcreport.tab_name.setText(str(index) + " - " + module[0])
      if module[2]["options"]["overwrite"] == True:
        self.qcreport.overwrite.setChecked(True)
      else:
        self.qcreport.overwrite.setChecked(False)
      if module[2]["options"]["skip"] == True:
        self.qcreport.skip.setChecked(True)
      else:
        self.qcreport.skip.setChecked(False)
      if module[2]["protocol"]["generatePDF"] == True:
        self.qcreport.generatePDF_true.setChecked(True)
      else: 
        self.qcreport.generatePDF_false.setChecked(True)
      if module[2]["protocol"]["generateCSV"] == True:
        self.qcreport.generateCSV_true.setChecked(True)
      else:
        self.qcreport.generateCSV_false.setChecked(False)
      self.details_stack.setCurrentIndex(11)

    self.communicate.SetUpdateModuleDetailsBool(True)



from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal as Signal


class ModuleSelectorCommunicate(QObject):

  set_drop_mode_protocol_list_widget = Signal()
  module_added_to_protocol_drop = Signal(list)
  call_execute_exclude_gradients_popup = Signal()
  call_execute_merge_images_popup = Signal()

  def SetDropModeProtocolListWidget(self):
    self.set_drop_mode_protocol_list_widget.emit()

  def ModuleAddedToProtocolDrop(self, list_modules):
    self.module_added_to_protocol_drop.emit(list_modules)

  def CallExecuteExcludeGradientsPopup(self):
    self.call_execute_exclude_gradients_popup.emit()

  def CallExecuteMergeImagesPopup(self):
    self.call_execute_merge_images_popup.emit()

class ModuleSelector(QWidget):

  communicate = ModuleSelectorCommunicate()

  def __init__(self, protocol_template):
    QWidget.__init__(self)
    self.protocol_template = protocol_template
    self.tab = QWidget()
    self.ModuleSelectorList()

  def ModuleSelectorList(self):
    modules_list_label = QLabel("Module Selector")
    self.modules_list_widget = QListWidget()    
    self.modules_list_widget.setSelectionMode(QAbstractItemView.ExtendedSelection)

    # setting drag drop mode
    self.modules_list_widget.setDragDropMode(QAbstractItemView.DragOnly)
    self.modules_list_widget.itemPressed.connect(self.DragDropMode)
    self.modules_list_widget.itemSelectionChanged.connect(self.ModulesMultiSelection)
    
    module_id = 0
    for module_iterator in self.protocol_template["options"]["execution"]["pipeline"]["candidates"]:
      if module_iterator["description"] != "Not implemented":
        new_module = QListWidgetItem()
        new_module.setText(module_iterator["caption"])
        new_module.setData(QtCore.Qt.UserRole, module_id)
        new_module.setToolTip(module_iterator["description"])
        module_id += 1
        self.modules_list_widget.addItem(new_module)

    layout_v = QVBoxLayout()
    layout_v.addWidget(modules_list_label)
    layout_v.addWidget(self.modules_list_widget)
    self.tab.setLayout(layout_v)

  def DragDropMode(self):
    self.communicate.SetDropModeProtocolListWidget()
    self.modules_list_widget.setDefaultDropAction(Qt.CopyAction)

  def ModulesMultiSelection(self):
    self.modules_list_widget.blockSignals(True)
    group_selection_with_special_module = False
    if len(self.modules_list_widget.selectedItems()) > 1:
      for module in self.modules_list_widget.selectedItems():
        if module.text() == "Exclude Gradients":
          group_selection_with_special_module = True
          self.communicate.CallExecuteExcludeGradientsPopup()
          break
        if module.text() == "Merge Images":
          group_selection_with_special_module = True
          self.communicate.CallExecuteMergeImagesPopup()
          break
      if group_selection_with_special_module:
        for module in self.modules_list_widget.selectedItems():
          module.setSelected(False)
    self.modules_list_widget.blockSignals(False)

  def GetListWidgetSelectedModules(self):
    modules_list = self.modules_list_widget.selectedItems()
    self.communicate.ModuleAddedToProtocolDrop(modules_list)

  def EnableModulesListWidget(self, new_state):
    self.modules_list_widget.setEnabled(new_state)

  def EnableExcludeGradientsModule(self, enable):
    if enable == True:
      self.modules_list_widget.findItems("Exclude Gradients", Qt.MatchExactly)[0].setFlags(self.modules_list_widget.findItems("Exclude Gradients", Qt.MatchExactly)[0].flags() | Qt.ItemIsEnabled)
    else:
      self.modules_list_widget.findItems("Exclude Gradients", Qt.MatchExactly)[0].setFlags(self.modules_list_widget.findItems("Exclude Gradients", Qt.MatchExactly)[0].flags() & ~Qt.ItemIsEnabled)

  def EnableMergeImagesModule(self, enable):
    if enable == True:
      self.modules_list_widget.findItems("Merge Images", Qt.MatchExactly)[0].setFlags(self.modules_list_widget.findItems("Merge Images", Qt.MatchExactly)[0].flags() | Qt.ItemIsEnabled)
    else:
      self.modules_list_widget.findItems("Merge Images", Qt.MatchExactly)[0].setFlags(self.modules_list_widget.findItems("Merge Images", Qt.MatchExactly)[0].flags() & ~Qt.ItemIsEnabled)

  def UpdateSelectorData(self, module_name, new_data):
    for module_iter in range(self.modules_list_widget.count()):
      if self.modules_list_widget.item(module_iter).text() == module_name:
        self.modules_list_widget.item(module_iter).setData(QtCore.Qt.UserRole, new_data)



#
# Reference : https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3864968/
#
# Written by SK Park , NIRAL, UNC
# 2021-04-18

  
import dtiplayground.dmri.preprocessing as prep

import yaml
from pathlib import Path

import SLICE_Check.computations as computations

logger=prep.logger.write

class SLICE_Check(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)

    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    @prep.measure_time
    def process(self,*args,**kwargs):  ## variables : self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        #logger(yaml.dump(inputParams))
        arte_sorted=computations.slice_check(self.image,computation_dir=self.computation_dir,
                                             headskip=self.protocol['headSkipSlicePercentage'],
                                             tailskip=self.protocol['tailSkipSlicePercentage'],
                                             baseline_z_Threshold=self.protocol['correlationDeviationThresholdbaseline'],
                                             gradient_z_Threshold=self.protocol['correlationDeviationThresholdgradient'],
                                             quad_fit=self.protocol['quadFit'],
                                             subregion_check=self.protocol['bSubregionalCheck'],
                                             subregion_relaxation_factor=self.protocol['subregionalCheckRelaxationFactor']
                                             )
        logger("-------------------------------------------------------------",prep.Color.WARNING)
        logger("Abnormal gradients",prep.Color.WARNING)
        logger("-------------------------------------------------------------",prep.Color.WARNING)

        grads=self.image.getGradients()
        for a in arte_sorted:
            grad_index=a[0]
            grad_original_index=grads[a[0]]['original_index']
            vec=grads[a[0]]['gradient']
            isB0=grads[a[0]]['baseline']
            logger("For gradient {} (Org.Idx {}) , Vec {}, isB0 {}".format(grad_index,
                                                                            grad_original_index,
                                                                            vec,
                                                                            isB0))
            for i in range(len(a[1])):
                logger("\t\tSlice.idx {}, Correlation : {:.4f}".format(a[1][i]['slice'],a[1][i]['correlation']))
        gradient_indexes_to_remove=[ix[0] for ix in arte_sorted]

        ## make result and set final image to self.result and self.image (which are to be copied to the next pipeline module as on input)
        ## Excluded original indexes will be automatically deleted in the postProcess
        #logger("Excluded gradient indexes : {}".format(gradient_indexes_to_remove),prep.Color.WARNING) #gradient indexes are not original one , so need to convert
        self.result['output']['excluded_gradients_original_indexes']=self.image.convertToOriginalGradientIndex(gradient_indexes_to_remove)
        #self.result['output']['image_path']=Path(self.output_dir).joinpath('output.nrrd').__str__()
        self.result['output']['success']=True
        self.image.setSpaceDirection(target_space=self.getSourceImageInformation()['space'])
        return self.result




import dtiplayground.dmri.preprocessing as prep
import dtiplayground.dmri.common.tools as tools
from dtiplayground.dmri.preprocessing.dwi import DWI
import yaml
from pathlib import Path
import importlib
###
import numpy as np
import ants 
import nibabel

logger=prep.logger.write

class BRAIN_Mask(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        
    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        # << TODOS>>
        protocol_options=args[0]
        self.num_threads=protocol_options['software_info']['parameters']['num_max_threads']
        self.software_info=protocol_options['software_info']['softwares']
        self.baseline_threshold=protocol_options['baseline_threshold']
        res=self.run_mask(method=self.protocol['method'],
                          modality=self.protocol['modality'],
                          averagingMethod=self.protocol['averagingMethod'])
        self.result['output']['success']=True
        return self.result

### User defined methods
    def mask_antspynet(self,params):
        import antspynet
        logger("AntsPyNet is running ...",prep.Color.INFO)
        res=None
        input_image_path=Path(self.output_dir).joinpath("input.nii.gz").__str__()
        output_mask_path=Path(self.output_dir).joinpath("mask.nii.gz").__str__()

        src_image=params['image']
        modality=params['modality']
        averagingMethod=params['averagingMethod']
        src_image.writeImage(input_image_path,dest_type='nifti')
        ants_image=ants.image_read(input_image_path)

        logger("Reduce to 3D volume for masking",prep.Color.INFO)
        baseline_img = self.image.extractBaselines()
        if len(baseline_img.getGradients()) < 1 :
            baseline_img = self.image
        reduced=baseline_img.reduceTo3D(method=averagingMethod)
        ants_image_3d=ants.from_numpy(data=reduced)
        ants_image_3d.set_origin(list(ants_image.origin)[:3])
        ants_image_3d.set_spacing(list(ants_image.spacing)[:3])
        new_dir=np.resize(ants_image.direction,[4,4])[:3,:3]
        ants_image_3d.set_direction(new_dir.tolist())
        logger("Computing probability mask ...",prep.Color.INFO)
        probability_mask=antspynet.brain_extraction(ants_image_3d,modality)
        logger("Generating thresholded mask ...",prep.Color.INFO)
        mask=ants.threshold_image(probability_mask,
                                  low_thresh=0.5,
                                  high_thresh=1.0,
                                  inval=1,
                                  outval=0)
        logger("Writing mask file",prep.Color.PROCESS)
        ants.image_write(mask, output_mask_path)
        dest_filename = input_image_path
        ## dev for nrrd output
        logger("Loading mask file",prep.Color.PROCESS)
        image=DWI(output_mask_path)
        # print(image.information)
        logger("Mask loaded",prep.Color.OK)
        logger("Saving NRRD")
        output_nrrd_path = Path(self.output_dir).joinpath("mask.nrrd").__str__()
        # print(image.information)
        image.setSpaceDirection(self.getSourceImageInformation()['space'])
        image.writeImage(output_nrrd_path,dest_type='nrrd')
        # print(image.information)
        logger("Saved as NRRD")
        ## dev-end
        self.addOutputFile(output_mask_path, 'Mask')
        self.addOutputFile(output_nrrd_path, 'Mask')
        self.addGlobalVariable('mask_path',output_nrrd_path)
        return res 

    def mask_fslbet(self,params): # provide baseline averaged image
        logger("FSL bet is running ...",prep.Color.INFO)
        input_image_base=Path(self.output_dir).joinpath("input").__str__()
        averaged_path=input_image_base+"_averaged"
        output_image_base=Path(self.output_dir).joinpath("mask").__str__()
        output_mask_base=Path(self.output_dir).joinpath("mask").__str__()

        input_image_path=input_image_base+".nii.gz"
        output_image_path=output_image_base+".nii.gz"
        output_mask_path=output_image_base+"_mask.nii.gz"
        output_mask_path_nrrd=output_image_base+".nrrd"
        src_image=params['image']
        averagingMethod=params['averagingMethod']
        src_image.writeImage(input_image_path,dest_type='nifti')
        fsl=tools.FSL(self.software_info['FSL']['path'])
        logger("Generating Mask",prep.Color.INFO)
        baseline_img = self.image.extractBaselines()
        if len(baseline_img.getGradients()) < 1 :
            baseline_img = self.image
        if averagingMethod == "direct_average":
            cmd_output=fsl.fslmaths_ops(input_image_path, averaged_path,'mean')
        if averagingMethod == "idwi":
            averaged_image=baseline_img.idwi() 
            affine_matrix=src_image.getAffineMatrixForNifti()           
            averaged_image = nibabel.Nifti1Image(averaged_image, affine=affine_matrix)
            nibabel.save(averaged_image, averaged_path)
        cmd_output=fsl.bet(averaged_path, output_image_path)
        mask=DWI(output_mask_path)
        mask.setSpaceDirection(self.getSourceImageInformation()['space'])
        mask.writeImage(output_mask_path_nrrd,dest_type='nrrd')
        self.addOutputFile(output_mask_path, 'Mask')
        self.addOutputFile(output_mask_path_nrrd, 'Mask')
        self.addGlobalVariable('mask_path',output_mask_path_nrrd)
        res=None
        return res

    def run_mask(self,method, modality,averagingMethod):
        res=None 
        params={}
        logger("Mask is being computed ... ",prep.Color.PROCESS)
        if method=='fsl':
            params={
                'image': self.image,
                'averagingMethod': averagingMethod
            }
            res=self.mask_fslbet(params)
        elif method=='antspynet':
            params={
                'image': self.image,
                'modality': modality,
                'averagingMethod': averagingMethod
            }
            res=self.mask_antspynet(params)
        logger("Mask generation is completed",prep.Color.OK)
        return res;





from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

class NoModule(QWidget):

  def __init__(self):
    QWidget.__init__(self)
    self.stack = QWidget()
    self.NoModuleStack()

  def NoModuleStack(self):
    label0 = QLabel("No module selected.")
    label1 = QLabel("Click on a module in the Protocol list to display its parameters.")
    label2 = QLabel("To add a module in the Protocol list, go to 'Select' tab, then select the module(s) of your choice and drag the module(s) into the Protocol list on the left.")
    label2.setWordWrap(True)

    no_module_layout_v = QVBoxLayout()
    no_module_layout_v.addStretch(1)
    no_module_layout_v.addWidget(label0)
    no_module_layout_v.addStretch(1)
    no_module_layout_v.addWidget(label1)
    no_module_layout_v.addWidget(label2)
    no_module_layout_v.addStretch(4)
    self.stack.setLayout(no_module_layout_v)


from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class SusceptibilityCorrection(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, susceptibility_yml, preferences_yml):
    QWidget.__init__(self)
    self.preferences_yml = preferences_yml
    self.stack = QWidget()
    self.SusceptibilityCorrectionStack(protocol_template, susceptibility_yml)

  def SusceptibilityCorrectionStack(self, protocol_template, susceptibility_yml):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "SUSCEPTIBILITY_Correct":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)   
    # overwrite
    self.overwrite = QCheckBox("Overwrite")
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    self.writeimage.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.writeimage, 0, 2)

    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QGridLayout()
    protocol_sublayout = QGridLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # phaseEncodingAxis
    self.phaseEncodingAxis_list = []
    protocol_layout.addWidget(QLabel("Phase Encoding Axis"), 0, 0, 1, 3)
    protocol_layout.addWidget(QLabel(susceptibility_yml["protocol"]["phaseEncodingAxis"]["caption"]), 0, 3, 1, 3)
    self.phaseEncodingAxis_p0 = QCheckBox("+0")
    self.phaseEncodingAxis_p1 = QCheckBox("+1")
    self.phaseEncodingAxis_p2 = QCheckBox("+2")

    self.phaseEncodingAxis_n0 = QCheckBox("-0") # for later implementation 
    self.phaseEncodingAxis_n1 = QCheckBox("-1")    
    self.phaseEncodingAxis_n2 = QCheckBox("-2")
    self.phaseEncodingAxis_n0.setEnabled(False)
    self.phaseEncodingAxis_n1.setEnabled(False)
    self.phaseEncodingAxis_n2.setEnabled(False)
    
    protocol_layout.addWidget(self.phaseEncodingAxis_n0, 1, 0)
    protocol_layout.addWidget(self.phaseEncodingAxis_p0, 1, 1)
    protocol_layout.addWidget(self.phaseEncodingAxis_n1, 1, 2)
    protocol_layout.addWidget(self.phaseEncodingAxis_p1, 1, 3)
    protocol_layout.addWidget(self.phaseEncodingAxis_n2, 1, 4)
    protocol_layout.addWidget(self.phaseEncodingAxis_p2, 1, 5)
    
    self.phaseEncodingAxis = QListWidget()
    self.phaseEncodingAxis.setDragDropMode(QAbstractItemView.InternalMove)
    self.phaseEncodingAxis.setSelectionMode(QAbstractItemView.ExtendedSelection)
    self.phaseEncodingAxis.model().rowsMoved.connect(self.UpdatePhaseEncodingAxisList)
    protocol_layout.addWidget(self.phaseEncodingAxis, 0, 6, 2, 1)
    self.phaseEncodingAxis_p0.stateChanged.connect(self.UpdatePhaseEncodingAxisWidget)
    self.phaseEncodingAxis_p0.stateChanged.connect(self.GetParams)
    self.phaseEncodingAxis_p1.stateChanged.connect(self.UpdatePhaseEncodingAxisWidget)
    self.phaseEncodingAxis_p1.stateChanged.connect(self.GetParams)
    self.phaseEncodingAxis_p2.stateChanged.connect(self.UpdatePhaseEncodingAxisWidget)
    self.phaseEncodingAxis_p2.stateChanged.connect(self.GetParams)
    # phaseEncodingValue
    self.phaseEncodingValue = QDoubleSpinBox()
    self.phaseEncodingValue.setAlignment(Qt.AlignRight)
    self.phaseEncodingValue.setSingleStep(0.001)
    self.phaseEncodingValue.setDecimals(3)
    self.phaseEncodingValue.setStatusTip(susceptibility_yml["protocol"]["phaseEncodingValue"]["description"])
    protocol_sublayout.addWidget(QLabel(susceptibility_yml["protocol"]["phaseEncodingValue"]["caption"]), 0, 0, 1, 1)
    protocol_sublayout.addWidget(self.phaseEncodingValue, 0, 1, 1, 4)
    self.phaseEncodingValue.valueChanged.connect(self.GetParams)
    # configurationFilePath
    self.configurationFilePath = QLineEdit()
    self.configurationFilePath.setStatusTip(susceptibility_yml["protocol"]["configurationFilePath"]["description"])
    protocol_sublayout.addWidget(QLabel(susceptibility_yml["protocol"]["configurationFilePath"]["caption"]), 1, 0, 1, 1)
    protocol_sublayout.addWidget(self.configurationFilePath, 2, 0, 1, 5)
    self.configurationFilePath_browse = QPushButton("Browse")
    protocol_sublayout.addWidget(self.configurationFilePath_browse, 1, 1, 1, 1)
    self.configurationFilePath.textChanged.connect(self.GetParams)
    self.configurationFilePath_browse.clicked.connect(self.BrowseFSLConfigFilePath)

    ## Layout
    for i in range(7):
      protocol_layout.setColumnStretch(i, 2)
    for i in range(4):
      protocol_layout.setRowStretch(i, 1)
    protocol_layout.addLayout(protocol_sublayout, 2, 0, 1, 7)
    
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

  def UpdatePhaseEncodingAxisWidget(self): # update QListWidget according to selected axis
    if self.phaseEncodingAxis_p0.isChecked():
      #if 0 not in list, then add it; else do nothing
      item_in_list = self.phaseEncodingAxis.findItems("0", Qt.MatchFixedString) # (QListWidget are not iterable)
      if not item_in_list:
        self.phaseEncodingAxis.addItem("0")
    else:
      #if 0 in list, then remove it; else do nothing
      item_in_list = self.phaseEncodingAxis.findItems("0", Qt.MatchFixedString)
      if item_in_list:
        self.phaseEncodingAxis.takeItem(self.phaseEncodingAxis.row(item_in_list[0]))
    
    if self.phaseEncodingAxis_p1.isChecked():
      #if 1 not in list, then add it; else do nothing
      item_in_list = self.phaseEncodingAxis.findItems("1", Qt.MatchFixedString)
      if not item_in_list:
        self.phaseEncodingAxis.addItem("1")      
    else:
      #if 1 in list, then remove it; else do nothing
      item_in_list = self.phaseEncodingAxis.findItems("1", Qt.MatchFixedString)
      if item_in_list:
        self.phaseEncodingAxis.takeItem(self.phaseEncodingAxis.row(item_in_list[0]))

    if self.phaseEncodingAxis_p2.isChecked():
      #if 2 not in list, then add it; else do nothing
      item_in_list = self.phaseEncodingAxis.findItems("2", Qt.MatchFixedString)
      if not item_in_list:
        self.phaseEncodingAxis.addItem("2") 
    else:
      #if 2 in list, then remove it; else do nothing
      item_in_list = self.phaseEncodingAxis.findItems("2", Qt.MatchFixedString)
      if item_in_list:
        self.phaseEncodingAxis.takeItem(self.phaseEncodingAxis.row(item_in_list[0]))
    self.UpdatePhaseEncodingAxisList()
    
  def UpdatePhaseEncodingAxisList(self): # update list according to QListWidget
    self.phaseEncodingAxis_list = []
    for ite in range(self.phaseEncodingAxis.count()):
      self.phaseEncodingAxis_list.append(int(self.phaseEncodingAxis.item(ite).text()))
    self.GetParams()

  def BrowseFSLConfigFilePath(self):
    file_filter = "Configuration file (*.cnf)"
    file_path = QFileDialog.getOpenFileName(
      parent = self,
      caption = "Select an configuration file",
      filter = file_filter
      )
    file_path = file_path[0]
    if file_path != "":
      self.preferences_yml["fslConfigurationFilePath"] = file_path
      self.configurationFilePath.setText(self.preferences_yml["fslConfigurationFilePath"])
      self.communicate.CallUpdateUserPreferencesFile(self.preferences_yml)

  def GetParams(self):
    params = [
      'Susceptibility correction',
      'SUSCEPTIBILITY_Correct', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        'write_image': self.writeimage.isChecked()
        }, 
      'protocol': {
        'configurationFilePath': self.configurationFilePath.text(), 
        'phaseEncodingAxis': self.phaseEncodingAxis_list, 
        'phaseEncodingValue': self.phaseEncodingValue.value()
        }
      } 
    ]
    self.communicate.SendParams(params)





  
import dtiplayground.dmri.preprocessing as prep

import yaml
from pathlib import Path 
import EDDYMOTION_Correct.utils as utils
import dtiplayground.dmri.common.tools as tools 
logger=prep.logger.write
from dtiplayground.dmri.common import measure_time
import shutil

logger=prep.logger.write

class EDDYMOTION_Correct(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)

    def generateDefaultEnvironment(self):
        return super().generateDefaultEnvironment()
    
    def checkDependency(self,environment): #use information in template, check if this module can be processed
        # FSL should be ready before execution
        software_path=Path(self.config_dir).joinpath('software_paths.yml')
        software_info = yaml.safe_load(open(software_path,'r'))
        if self.name in environment:
            fslpath=Path(software_info['softwares']['FSL']['path'])
            try:
                #fslpath=Path(environment[self.name]['fsl_path'])
                fsl_exists=fslpath.exists()
                if fsl_exists:
                    return True, None 
                else:
                    return False, "FSL Path doesn't exist : {}".format(str(fslpath))
            except Exception as e:
                return False, "Exception in finding FSL6 : {}".format(str(fslpath))
        else:
            return False, "Can't locate FSL" #test

    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.global_variables, self.softwares, self.output_dir, self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        gradient_indexes_to_remove=[]
        protocol_options=args[0]
        self.num_threads=protocol_options['software_info']['parameters']['num_max_threads']
        self.software_info=protocol_options['software_info']['softwares']
        susceptibility=False
        susceptibility_parameters=None
        self.qcReport = self.protocol['qcReport']
        if 'susceptibility_parameters' in inputParams:
            susceptibility=True
            susceptibility_parameters=inputParams['susceptibility_parameters']
        res=None
        if susceptibility:
            res=self.eddy(self.image,
                          outfilename=None,
                          params=susceptibility_parameters,
                          protocols=self.protocol)
        else:
            res=self.eddy(self.image,
                          outfilename=None,
                          params=None,
                          protocols=self.protocol)

        ## results
        self.result['output']['excluded_gradients_original_indexes']=self.image.convertToOriginalGradientIndex(gradient_indexes_to_remove)
        self.result['output']['success']=True
        return self.result

### User defined methods
    ### fsl parameters
    def make_acqp(self,axis=0,val=0.0924):
        acqps=[[0,1,0,val]]
        outfilename=Path(self.output_dir).joinpath('acqp.txt').__str__()
        with open(outfilename,'w') as f:
            for e in acqps:
                strline="{:d} {:d} {:d} {:.4f}\n".format(*e)
                f.write(strline)
        return outfilename

    def make_index(self,image):
        outfilename=Path(self.output_dir).joinpath('index.txt').__str__()
        grads=image.getGradients()
        with open(outfilename,'w') as f:
            for g in grads:
                f.write("1 ")
        return outfilename 

    def make_index_new(self,image,baseline_index_file):
        grads=image.getGradients()
        bidx_sarr=open(baseline_index_file,'r').read().split()
        bidx=list(map(int,bidx_sarr))
        index_filename=Path(self.output_dir).joinpath('index.txt').__str__()

        b0idx=0
        res_index=[]
        for i,g in enumerate(grads):
            if b0idx+1 < len(bidx):
                if i>=bidx[b0idx+1]:
                    b0idx+=1 
            res_index.append(b0idx+1)
        res_str=" ".join(list(map(str,res_index)))
        open(index_filename,'w').write(res_str)
        return index_filename

    ### scripts

    @measure_time
    def eddy(self,image,outfilename,params, protocols): ## eddy with topup (susceptibility correction process is required before execution)

        output_dir=Path(self.output_dir)
        input_nifti=output_dir.joinpath('input.nii.gz').__str__()
        input_bvals=output_dir.joinpath('input.bval').__str__()
        input_bvecs=output_dir.joinpath('input.bvec').__str__()
        output_nifti=output_dir.joinpath('output.nii.gz').__str__()
        binary_mask=output_dir.joinpath('output_mask.nii.gz').__str__()
        ### conversion to nifti 
        if params is not None: ## if susceptibility parameters present
            input_nifti=params['image_path']
            input_bvals=params['image_bvals_path']
            input_bvecs=params['image_bvecs_path']
            binary_mask=params['mask_path']
        processed_nifti=output_dir.joinpath('output_eddied.nii.gz').__str__()
        processed_nifti_base=Path(processed_nifti).parent.joinpath(Path(processed_nifti).name.split('.')[0]).__str__()
        processed_nifti_nonneg=output_dir.joinpath('output_eddied_nonneg.nii.gz').__str__()
        processed_bvals=output_dir.joinpath('output_eddied.bval').__str__()
        processed_bvecs=output_dir.joinpath('output_eddied.bvec').__str__()
        output_nrrd=output_dir.joinpath('output.nrrd').__str__()
        quad_output_dir=output_dir.joinpath('output_eddied.qc').__str__()

        ### generate mask
        fsl=tools.FSL(self.software_info['FSL']['path'])
        fsl._set_num_threads(self.num_threads)
        fsl.setDevMode(True)

        ### acqp file writing
        acqp_filename=None
        index_filename=None
        topup_filename=None
        _average_path=output_dir.joinpath("_average.nii.gz").__str__()
        img=None
        if params is not None: ##susceptibility case
            acqp_filename=params['acqp_path']
            index_filename=self.make_index_new(self.image,params['index_path'])
            topup_filename=params['topup_path']
        else: # single eddy
            self.writeImageWithOriginalSpace(str(input_nifti),dest_type='nifti')
            img=self.loadImage(input_nifti)
            logger("Generating Mask : {}".format(binary_mask.__str__()))
            output=fsl.fslmaths_ops(input_nifti,_average_path,'mean')
            res=fsl.bet(str(input_nifti),str(output_nifti))
            ouput_nifti=Path(output_nifti).rename(output_dir.joinpath('temp.nii.gz').__str__())
            acqp_filename=self.make_acqp() 
            index_filename=self.make_index(self.image)
        ### eddy correction
        if not Path(processed_nifti).exists():
            if params is not None:
                logger("Computing eddy with susceptibility correction... ",prep.Color.PROCESS)
                res=fsl.eddy_openmp(imain=input_nifti,
                                mask=binary_mask,
                                acqp=acqp_filename,
                                index_file=index_filename,
                                bvals=input_bvals,
                                bvecs=input_bvecs,
                                out=processed_nifti_base, #basename
                                estimate_move_by_susceptibility=protocols['estimateMoveBySusceptibility'],
                                topup=topup_filename,
                                data_is_shelled=protocols['dataIsShelled'],
                                repol=protocols['interpolateBadData'])
            else:
                logger("Computing eddy ... ",prep.Color.PROCESS)
                res=fsl.eddy_openmp(imain=input_nifti,
                                mask=binary_mask,
                                acqp=acqp_filename,
                                index_file=index_filename,
                                bvals=input_bvals,
                                bvecs=input_bvecs,
                                out=processed_nifti_base,
                                estimate_move_by_susceptibility=False,
                                topup=None,
                                data_is_shelled=protocols['dataIsShelled'],
                                repol=protocols['interpolateBadData'])
        else:
            logger("Eddymotion corrected output exists: {}".format(processed_nifti),prep.Color.OK)
            self.image=self.loadImage(processed_nifti)
        shutil.copy(processed_nifti_base+".eddy_rotated_bvecs",processed_bvecs)
        shutil.copy(input_bvals,processed_bvals)

        img=self.loadImage(processed_nifti)
        if not Path(output_dir.joinpath("output_eddied_dev.nrrd")).exists():
            img.writeImage(Path(output_dir.joinpath("output_eddied_dev.nrrd")).__str__(),dest_type='nrrd')

        logger("Generating Non negative DWI...",prep.Color.PROCESS)
        nonneg_base=Path(processed_nifti_nonneg).name.split('.')[0]
        processed_nifti_nonneg_bvals=Path(processed_nifti_nonneg).parent.joinpath(nonneg_base+".bval").__str__()
        processed_nifti_nonneg_bvecs=Path(processed_nifti_nonneg).parent.joinpath(nonneg_base+".bvec").__str__()

        if not Path(processed_nifti_nonneg).exists():
            output=fsl.fslmaths_threshold(processed_nifti,processed_nifti_nonneg,0)
            shutil.copy(processed_bvals,processed_nifti_nonneg_bvals)
            shutil.copy(processed_bvecs,processed_nifti_nonneg_bvecs)

        img=self.loadImage(processed_nifti_nonneg)
        if not Path(output_dir.joinpath("output_eddied_nonneg_dev.nrrd")).exists():
            img.writeImage(Path(output_dir.joinpath("output_eddied_nonneg_dev.nrrd")).__str__(),dest_type='nrrd')

        if not Path(quad_output_dir).exists() and protocols['qcReport']:
            logger("Executing eddy_quad for quality assessment...",prep.Color.PROCESS)
            output=fsl.eddy_quad(
                        input_base=processed_nifti_base,
                        idx=index_filename,
                        par=acqp_filename,
                        mask=binary_mask,
                        bvals=processed_bvals)

        self.image=self.loadImage(processed_nifti_nonneg)
        self.image.image_type='nrrd'
        self.writeImageWithOriginalSpace(output_nrrd,'nrrd')
        return None




from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 
import re

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class ExcludeGradients(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, exclude_yml):
    QWidget.__init__(self)
    self.stack = QWidget()

    self.gradientsToExclude_list = []

    self.ExcludeGradientsStack(protocol_template, exclude_yml)

  def ExcludeGradientsStack(self, protocol_template, exclude_yml):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "MANUAL_Exclude":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)   
    # overwrite
    self.overwrite = QCheckBox("Overwrite")
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    self.writeimage.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.writeimage, 0, 2)
    
    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QVBoxLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # selectionMethod
    self.selectionMethod = QComboBox()
    self.selectionMethod.addItem("Manual Selection")
    self.selectionMethod.addItem("QuickView")
    self.selectionMethod.setStatusTip("Method used to select the gradients to exclude")
    self.selectionMethod.currentTextChanged.connect(self.EnableDisable)  
    protocol_layout.addWidget(self.selectionMethod)
    self.selectionMethod.currentTextChanged.connect(self.GetParams)
    # gradientsToExclude
    protocol_layout.addWidget(QLabel(exclude_yml["protocol"]["gradientsToExclude"]["caption"] + " (comma separated list)"))
    self.gradients2exclude = QTextEdit()
    self.gradients2exclude.setStatusTip(exclude_yml["protocol"]["gradientsToExclude"]["description"])
    self.gradients2exclude.textChanged.connect(self.GetGradients)
    protocol_layout.addWidget(self.gradients2exclude)
    self.gradients2exclude.textChanged.connect(self.GetParams)

    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

  def GetGradients(self): # get list of gradients typed by user
    self.gradientsToExclude_list = []
    text = self.gradients2exclude.toPlainText()
    text = text.split(",")
    for elmt in text:
      elmt = re.sub(r"\s+", "", elmt)
      if elmt.isnumeric():
        self.gradientsToExclude_list.append(int(elmt))    

  def EnableDisable(self): # disable space to write gradient indexes if QuickView is selected
    if self.selectionMethod.currentText() != "Manual Selection":
      self.gradients2exclude.setEnabled(False)
    else:
      self.gradients2exclude.setEnabled(True)

  def GetParams(self):
    params = [
      'Exclude Gradients',
      'MANUAL_Exclude', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        'write_image': self.writeimage.isChecked()
        }, 
      'protocol': {
        'gradientsToExclude': self.gradientsToExclude_list, 
        }
      }
    ]
    self.communicate.SendParams(params)


from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class BRAINSFit(ExternalToolWrapper):
    def __init__(self,binary_path):
        super().__init__(binary_path)

    
    def affine_registration(self,
                            fixed_path,
                            moving_path,
                            output_path,
                            output_transform_path,
                            initial_transform_path,
                            transform_mode):
        arguments=['--fixedVolume',fixed_path,
                           '--movingVolume',moving_path,
                           '--useAffine',
                           '--outputVolume',output_path,
                           '--outputTransform',output_transform_path
                           ]
        if initial_transform_path is None:
            arguments+=['--initializeTransformMode', transform_mode]
        else:
            arguments+=['--initialTransform',initial_transform_path]
        self.setArguments(arguments)
        return self.execute(arguments)



from dtiplayground.dmri.common.tools.base import ExternalToolWrapper

class DTIAverage(ExternalToolWrapper):
    def __init__(self,binary_path):
        super().__init__(binary_path)

    
    def average(self,input_file_list:list,output_file,options=[]):
        arguments=[]
        for fn in input_file_list:
            arguments+=['--inputs',fn]
        arguments+=['--tensor_output',output_file]
        arguments+=options
        self.setArguments(arguments)
        return self.execute(arguments)
        



INFO = {
  "dtiplayground": { 
    "version" : "0.3.7"
  },
  "dmriprep": {
    "version" : "0.3.7"

  },
  "dtiatlasbuilder": {
    "version" : "0.1.0"
  },
  "autotract" : {
    "version" : "0.0.1"
  }
}




class DMRIXInput(object):  ## standard data
  def __init__(*args,**kwargs):
    self.inputs = []
    pass


class DMRIPlayground(object):
  def __init__(*args,**kwargs):
    pass


  def initialize(self):
    pass

  def preprocess(self):
    pass

  def process(self):
    pass

  def postprocess(self):
    pass


import dtiplayground.dmri.preprocessing as prep

import numpy as np
import time
from pathlib import Path
import yaml
import os 

logger=prep.logger.write


def ncc_old(x,y):  #normalized correlation
    x=(x-np.mean(x))/np.sqrt(np.sum((x-np.mean(x))**2))
    y=(y-np.mean(y))/np.sqrt(np.sum((y-np.mean(y))**2))
    return np.sum(x*y)

def ncc(x,y): #normalized cross correlation in image (ref: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3864968/ )
    ab=np.sum(x*y)
    a2=np.sum(x**2)
    b2=np.sum(y**2)
    if a2*b2==0.0: 
        return 1.0
    else:
        return ab/np.sqrt(a2*b2)

def _quad_fit(bval, domain=[0,1000], fimage=[3.0,3.5]): #returns std multiple between fimage
    if bval > domain[1] : 
        return fimage[1]
    elif bval < domain[0] : 
        return fimage[0]
    denom=((domain[1]-domain[0])**2)
    a=0
    if denom==0: 
        a=0
    else:
        a= (fimage[1]-fimage[0])/denom
    c= fimage[0]
    return a*(bval**2)+c

def quadratic_fit_generator(domain,fimage):
    def wrapper(bval):
        return _quad_fit(bval,domain,fimage)
    return wrapper

@prep.measure_time
def slice_check(image, computation_dir, # image object containing all the information
                headskip=0.1, 
                tailskip=0.1, 
                baseline_z_Threshold= 3.0, 
                gradient_z_Threshold=3.5, 
                quad_fit=True , 
                subregion_check=False, ## not implemented yet
                subregion_relaxation_factor=1.1):
    
    image_tensor=image.images.astype(float) 
    gradients=image.getGradients()

    ## Generate slice correlation informations over gradients
    gsum=[]
    begin_slice=int(image_tensor.shape[2]*headskip)
    last_slice=int(image_tensor.shape[2]*(1-tailskip))
    for k in range(image_tensor.shape[3]):
        csum=[]
        for i in range(image_tensor.shape[2]):
            if i<=begin_slice or i>=last_slice: continue
            af=image_tensor[:,:,i,k].reshape([1,-1])
            bf=image_tensor[:,:,i-1,k].reshape([1,-1])
            corr=ncc(af,bf)
            if not np.isnan(corr): csum.append(float(corr))
            else: csum.append(0.0)
        gsum.append(csum)
    gsum=np.array(gsum) 

    ## lookup artifacts from the z-threshold criteria over gradients for each slices.
    min_bval=min(list(map(lambda x: x['b_value'],gradients)))
    max_bval=max(list(map(lambda x: x['b_value'],gradients)))
    quadfit= quadratic_fit_generator([min_bval,max_bval],[baseline_z_Threshold,gradient_z_Threshold])
    artifacts={}
    for slice_index in range(gsum.shape[1]):
        gradwise_vec=gsum[:,slice_index]
        avg=np.mean(gradwise_vec)
        std=np.std(gradwise_vec)
        logger("Slice {}, Mean {:.4f}, Std {:.4f}".format(slice_index+begin_slice,avg,std))
        for idx,g in enumerate(gradwise_vec):
            z_threshold=baseline_z_Threshold
            if quad_fit:
                z_threshold=quadfit(gradients[idx]['b_value'])
            elif not gradients[idx]['baseline']:
                z_threshold=gradient_z_Threshold
            if g < avg- z_threshold*std:
                if idx not in artifacts: artifacts[idx]=[]
                artifacts[idx].append({"slice":slice_index+begin_slice,
                                        "correlation":float(g),
                                        'z_threshold':float(z_threshold),
                                        'z_value':float((g-avg)/std),
                                        'b_value':float(gradients[idx]['b_value'])})

    gsum_file=Path(computation_dir).joinpath('correlation_table.yml') # row=gradient index, col = slice index
    yaml.dump(gsum.tolist(),open(gsum_file,'w'))
    artifacts_file=Path(computation_dir).joinpath("artifacts.yml")
    yaml.dump(artifacts,open(artifacts_file,'w'))

    ## recap and return the results
    arte=list(artifacts.items())
    arte_sorted=sorted(arte,key=lambda x: x[0]) ## fitst element is gradient index, second is list of slice indexes with artifacts
    return arte_sorted





import dtiplayground.dmri.preprocessing as prep
import yaml
from pathlib import Path

###
import numpy as np

logger=prep.logger.write

class TEST(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        
    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.global_variables, self.softwares, self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        sourceImageInformation=self.getSourceImageInformation()
        # << TODOS>>
        protocol_options=args[0]
        self.num_threads=protocol_options['software_info']['parameters']['num_max_threads']
        self.software_info=protocol_options['software_info']['softwares']
        self.baseline_threshold=protocol_options['baseline_threshold']
        res=self.run_test(testlist=self.protocol['test_list'])
        self.result['output']['success']=True
        return self.result


    def run_test(self,testlist):
        res=None 
        params={}
        logger("Test is running... ",prep.Color.PROCESS)
        params={
            "image":self.image
        }
        num_test=len(testlist)
        for idx,test in enumerate(testlist):
            logger("-----------------------------------------\n{}/{}  {} is running ...\n-----------------------------------------".format(idx+1,num_test,test['name']))
            logger("------- OPTIONS-------\n{}".format(yaml.dump(test['options'])),prep.Color.DEV)
            func=getattr(self,test['name'])
            res=func(params,test['options'])
        logger("Test is completed",prep.Color.OK)
        return res;

### test functions 

    def image_io_test(self,params,options):
        
        res=None
        src_image=params['image']
        input_image_path=Path(self.output_dir).joinpath("input.nii.gz").__str__()
        src_image.writeImage(input_image_path,dest_type='nifti')
        temp_nrrd_path=Path(self.output_dir).joinpath("input_directsave.nrrd").__str__()
        src_image.writeImage(temp_nrrd_path,dest_type='nrrd')
        temp_path=Path(self.output_dir).joinpath("input_resave.nii.gz").__str__()
        img=self.loadImage(input_image_path)
        img.writeImage(temp_path)
        temp_nrrd_path=Path(self.output_dir).joinpath("input_resave.nrrd").__str__()
        img.writeImage(temp_nrrd_path,dest_type='nrrd')

        return res 

    def global_variable_test(self, params, options):
        print(yaml.dump(self.global_variables))
        return True





import dtiplayground.dmri.preprocessing as prep

import yaml, inspect
from pathlib import Path 
import pkgutil,sys, copy
import os

logger=prep.logger.write

@prep.measure_time
def _load_modules(user_module_paths=[],module_names=None, **options):
    modules=_load_default_modules(module_names, **options)
    usermodules= _load_modules_from_paths(user_module_paths,module_names, **options)
    modules.update(usermodules)
    return modules

def _load_default_modules(module_names=None, **options): # user_modules list of paths of user modules
    modules={}
    default_module_paths=[Path(__file__).resolve().parent]
    return _load_modules_from_paths(default_module_paths,module_names, **options)

def _load_modules_from_paths(user_module_paths: list,module_names=None, **options): #module names = list of modules to load (if none, load everything)
    modules={}
    mods=[]
    for pth in map(lambda x: str(x),user_module_paths):  ## path objects to string array
        logger("Loading modules from {} ".format(str(pth)),prep.Color.PROCESS)
        sys.path.insert(0, pth)
        pkgs_info=list(pkgutil.walk_packages([pth]))
        for p in pkgs_info:
            if module_names is not None:
                if p.name.split('.')[0] in module_names:
                    if len(p.name.split('.'))==1:
                        logger("Loading module : {}".format(p.name),prep.Color.OK)
                    mods.append(p.module_finder.find_module(p.name).load_module(p.name))
            else:
                if len(p.name.split('.'))==1:
                    logger("Loading module : {}".format(p.name),prep.Color.OK)
                mods.append(p.module_finder.find_module(p.name).load_module(p.name))
                    
        mod_filtered=list(filter(lambda x: len(x.__name__.split('.'))==2 and x.__name__.split('.')[0]==x.__name__.split('.')[1] ,mods))
        
        for md in mod_filtered:
            fn=Path(md.__file__)
            template_path= fn.parent.joinpath(fn.stem+'.yml')
            template=yaml.safe_load(open(template_path,'r'))
            name=md.__name__.split('.')[0]  #module name
            modules[name]={
                                "name" : name,
                                "module" : md,
                                "path" : md.__file__,
                                "template" : template,
                                "template_path" : template_path,
                                "valid" : False,
                                "validity_message" : None,
                                "options" : options
                                } 
    return modules 

@prep.measure_time
def check_module_validity(modules:list,environment,config_dir):
    logger("Checking dependencies ...",prep.Color.PROCESS)
    for name,md in modules.items():
                validity, msg=getattr(md['module'], name)(config_dir).checkDependency(environment)
                if not validity:
                    logger("[WARNING] Dependency is not met for the module : {} , {}".format(name,msg),prep.Color.WARNING)
                modules[name]['valid']=validity
                modules[name]['validity_message']=msg
    logger("Checking dependencies DONE",prep.Color.OK)
    return modules 


def generate_module_envionrment(modules :list,install_dir,*args,**kwargs):
    env={}
    for name,m in modules.items():
        mod_instance=getattr(m['module'],name)(install_dir)
        mod_instance.install(install_dir,*args,**kwargs)
        module_env=mod_instance.generateDefaultEnvironment()
        env[name]=module_env
    return env 

load_modules=_load_modules

def empty_result():
    res={
            "module_name" : None,
            "input" : None,
            "output": {
                "image_path" : None, #output image path (string)
                "image_object" : None,
                "output_directory": None,
                "excluded_gradients_original_indexes": [],
                "output_path": None,
                "success" : False,
                "image_information": None,
                "global_variables": {}
            } 
        }

    return res 

class DTIPrepModule: #base class
    def __init__(self,config_dir,*args, **kwargs):
        self.name=self.__class__.__name__
        self.config_dir=config_dir
        self.source_image=None
        self.image=None #image for output
        self.images=[] #image list for the multi-input modules
        self.image_paths=[] 
        self.protocol=None
        self.result_history=None
        self.result=empty_result()
       
        ##
        self.template=None ## template information located at the same directory of this module.(modulename.yml)
        self.output_dir=None ## output root
        self.output_root=None ## output root (one level up)
        self.computation_dir=None ## computation files 
        self.options={} ## this is pipeline option, not protocol. It should not affect the way of computing, only does to the behaviour of execution
        self.output_files = []
        self.software_info = {}
        self.softwares = {}
        self.global_variables={}
        if 'software_info' in kwargs:
            self.software_info= kwargs['software_info']
            self.softwares = self.software_info['softwares']
        if 'global_variables' in kwargs:
            self.global_variables.update(kwargs['global_variables'])
        ## loading template file (yml)
        self.loadTemplate()

    @prep.measure_time
    def initialize(self,result_history,image_path,output_dir):
        self.history=result_history
        self.result_history=result_history[image_path]
        self.output_dir=output_dir
        self.output_root=str(Path(self.output_dir).parent.parent)
        self.computation_dir=Path(output_dir).joinpath("computations")
        self.computation_dir.mkdir(parents=True,exist_ok=True)
        inputpath=None

        if 'multi_input' in self.template['process_attributes']:
            previous_outputs=[]
            for k,v in self.history.items():
                if k != image_path:
                    previous_outputs.append(v[-1])
            
            for idx,previous_result in enumerate(previous_outputs):
                if previous_result["output"]["image_object"] is not None:
                    self.source_image=prep.object_by_id(previous_result["output"]["image_object"])
                    self.images.append(copy.deepcopy(self.source_image))
                    logger("Source Image (Previous output) loaded from memory (object id): {}".format(id(self.source_image)),prep.Color.OK)
                else:
                    logger("Loading image from the file : {}".format(previous_result['output']['image_path']),prep.Color.PROCESS)
                    src_image_filename=Path(self.output_root).joinpath(previous_result['output']['image_path']).__str__()
                    self.source_image=self.loadImage(src_image_filename)
                    self.images.append(copy.deepcopy(self.source_image))
                        ## gradient information update
                    prev_output_dir=Path(self.output_root).joinpath(previous_result['output']['output_directory'])
                    prev_gradient_filename=prev_output_dir.joinpath('output_gradients.yml').__str__()
                    prev_image_information_filename=prev_output_dir.joinpath('output_image_information.yml').__str__()
                    self.images[idx].loadGradients(prev_gradient_filename)
                    self.images[idx].loadImageInformation(prev_image_information_filename)
                    logger("Source Image (Previous output) loaded",prep.Color.OK)

                    ### dump initial gradients and image information
                gradient_filename=str(Path(self.output_dir).joinpath('input_gradients_{:02d}.yml'.format(idx)))
                image_information_filename=str(Path(self.output_dir).joinpath('input_image_information_{:02d}.yml'.format(idx)))

                self.images[idx].dumpGradients(gradient_filename)
                self.images[idx].dumpInformation(image_information_filename)

            self.result_history.append({"output":previous_outputs})
            self.image=copy.deepcopy(self.images[0])
            self.result["module_name"]=self.name 
            self.result["input"]=previous_outputs
            self.result["output"]["image_object"]= id(self.image)
            self.result["output"]["success"]=False 
            self.result["output"]["output_directory"]=str(Path(self.output_dir).relative_to(self.output_root))

        else:
            #inputpath=Path(self.result_history[0]["output"]["image_path"]).absolute()
            previous_result=self.getPreviousResult()
            logger(yaml.safe_dump(previous_result))
            if previous_result["output"]["image_object"] is not None:
                self.source_image=prep.object_by_id(previous_result["output"]["image_object"])
                self.image=copy.deepcopy(self.source_image)
                logger("Source Image (Previous output) loaded from memory (object id): {}".format(id(self.source_image)),prep.Color.OK)
            else:
                logger("Loading image from the file : {}".format(previous_result['output']['image_path']),prep.Color.PROCESS)
                src_image_filename=Path(self.output_root).joinpath(previous_result['output']['image_path']).__str__()
                self.source_image=self.loadImage(src_image_filename)
                self.image=copy.deepcopy(self.source_image)
                    ## gradient information update
                prev_output_dir=Path(self.output_root).joinpath(previous_result['output']['output_directory'])
                prev_gradient_filename=prev_output_dir.joinpath('output_gradients.yml').__str__()
                prev_image_information_filename=prev_output_dir.joinpath('output_image_information.yml').__str__()
                self.image.loadGradients(prev_gradient_filename)
                self.image.loadImageInformation(prev_image_information_filename)
                logger("Source Image (Previous output) loaded",prep.Color.OK)

            ### dump initial gradients and image information
            gradient_filename=str(Path(self.output_dir).joinpath('input_gradients.yml'))
            image_information_filename=str(Path(self.output_dir).joinpath('input_image_information.yml'))

            self.image.dumpGradients(gradient_filename)
            self.image.dumpInformation(image_information_filename)

            self.result["module_name"]=self.name 
            self.result["input"]=previous_result["output"]
            self.result["output"]["image_object"]= id(self.image)
            self.result["output"]["success"]=False 
            self.result["output"]["output_directory"]=str(Path(self.output_dir).relative_to(self.output_root))


    def install(self,install_dir=None,*args,**kwargs):
        pass 

    def checkDependency(self,environment={}): #use information in template, check if this module can be processed
        return True , None

    def addOutputFile(self, src_filename, postfix):
        self.output_files.append({'source': src_filename, 'postfix': postfix})

    def addGlobalVariable(self, key, value):
        self.result['output']['global_variables'][key]=value

    def updateGlobalVariable(self, global_vars):
        self.result['output']['global_variables'].update(global_vars)
        
    def getGlobalVariables(self):
        return self.result['output']['global_variables']

    def getOutputFiles(self):
        return self.output_files

    @prep.measure_time
    def writeImage(self,filename,dest_type='nrrd',dtype='short'):
        self.image.writeImage(filename,dest_type=dest_type,dtype=dtype)
        #self.result['output']['image_path']=str(Path(filename).absolute().relative_to(self.output_root))
        self.result['output']['image_path']=str(Path(filename).absolute())

    @prep.measure_time
    def writeImageWithOriginalSpace(self,filename,dest_type='nrrd',dtype='short'):
        target_space = self.getSourceImageInformation()['space']
        self.image.setSpaceDirection(target_space=target_space)
        self.writeImage(filename,dest_type,dtype)

    @prep.measure_time 
    def loadImage(self, image_path, gradient_path=None):
        grad_path=gradient_path 
        if gradient_path is None:
            grad_path=Path(image_path).parent.joinpath('gradients.yml')

        image=prep.dwi.DWI(image_path)
        if grad_path.exists():
            grad=yaml.safe_load(open(grad_path,'r'))
            image.setGradients(grad)
        return image

    def getInputResult(self):
        return self.result_history[0]

    def getPreviousResult(self):
        return self.result_history[-1]

    def getSourceImageInformation(self):
        inputInfo = self.getInputResult()
        info=None
        #if 'multi_input' in self.template['process_attributes']:
        if type(inputInfo['output'])==list:
            info = inputInfo['output'][0]['output']['image_information']
        else:
            info = inputInfo['output']['image_information']
        return info

    def loadTemplate(self):
        modulepath=inspect.getfile(self.__class__)
        template_filename=Path(modulepath).parent.joinpath(self.name+".yml")
        self.template=yaml.safe_load(open(template_filename,'r'))

    def setImage(self, image ):
        self.image=image

    def setOptionsAndProtocol(self,options):
        self.options=options['options']
        self.protocol=options['protocol']

    def setProtocol(self,options):
        self.protocol=options['protocol']

    def getProtocol(self):
        return self.protocol

    def setOptions(self,options):
        self.options=options['options']

    def getOptions(self):
        return self.options 

    def getTemplate(self):
        return self.template 

    def generateDefaultProtocol(self,image_obj):
        self.protocol={}
        for k,v in self.template['protocol'].items():
                self.protocol[k]=v['default_value']
        return self.protocol

    def generateDefaultEnvironment(self):
        return None


    def process(self,*args,**kwargs): ## returns new result array (User implementation), returns output result
        #anything common
        ## variables : self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        pass

    @prep.measure_time
    def postProcess(self,result_obj,opts):
        self.result=result_obj
        if "multi_input" in self.template['process_attributes']:
            self.image=self.loadImage(self.result['output']['image_path'])
            self.result['output']['image_object']=id(self.image)
        self.result['input']=self.getPreviousResult()['output']
        self.image.deleteGradientsByOriginalIndex(self.result['output']['excluded_gradients_original_indexes'])
        logger("Excluded gradient indexes (original index) : {}"
            .format(self.result['output']['excluded_gradients_original_indexes']),prep.Color.WARNING)

        gradient_filename=str(Path(self.output_dir).joinpath('output_gradients.yml'))
        image_information_filename=str(Path(self.output_dir).joinpath('output_image_information.yml'))
        self.result['output']['image_object']=id(self.image)
        ### if deform_image process type, then forced file loading should be done from the second run in the subsequent process.
        if "deform_image" in self.template['process_attributes']:        
            if  Path(self.output_dir).joinpath('result.yml').exists() and not self.options['overwrite']: ## second run
                self.result['output']['image_object']=None 
                #self.image.gradients=yaml.safe_load(open(gradient_filename,'r'))
                self.image.loadGradients(gradient_filename)
                self.image.loadImageInformation(image_information_filename)
                #self.image.information=yaml.safe_load(open(image_information_filename,'r'))
            else: ## first run
                self.result['output']['image_object']=id(self.image)
        ### deform_image ends

        ### set baseline_threahold
        baseline_threshold=opts['baseline_threshold']
        self.image.setB0Threshold(baseline_threshold)
        self.image.getGradients()

        self.result['output']['success']=True
        self.result['output']['image_information']=self.image.information
        outstr=yaml.safe_dump(self.result)
        with open(str(Path(self.output_dir).joinpath('result.yml')),'w') as f:
            yaml.safe_dump(self.result,f)
        self.image.dumpGradients(gradient_filename)
        self.image.dumpInformation(image_information_filename)

        ## output gradients summary
        b_grads, _ =self.image.getBaselines()
        grad_summary = self.image.gradientSummary()
        logger("Remaining Gradients summary - Num.Gradients: {}, Num.Baselines: {}"
            .format(grad_summary['number_of_gradients'],
                    grad_summary['number_of_baselines']),prep.Color.INFO)

        logger("Remaining baselines",prep.Color.INFO)
        for g in b_grads:
            logger("[Gradient.idx {:03d} Original.idx {:03d}] Gradient Dir {} B-Value {:.1f}"
                .format(g['index'],g['original_index'],g['gradient'],g['b_value']),prep.Color.INFO)


    @prep.measure_time
    def run(self,*args,**kwargs): #wrapper 
        opts=args[0]
        baseline_threshold=opts['baseline_threshold']
        if 'global_vars' in kwargs:
            self.result['output']['global_variables'].update(kwargs['global_vars'])
            self.global_variables.update(kwargs['global_vars'])
        self.image.setB0Threshold(baseline_threshold)
        self.image.getGradients()
        
        res=self.process(*args,**kwargs) ## main computation for user implementation
        ## Post processing
        self.postProcess(res,opts) ## pretty much automatic        
        #return self.result["output"]["success"]
        return self.result["output"]

    def getResultHistory(self):
        return self.result_history+[self.result]


from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class InterlaceCheck(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, interlacecheck_yml):
    QWidget.__init__(self)
    self.stack = QWidget()
    self.InterlaceCheckStack(protocol_template, interlacecheck_yml)

  def InterlaceCheckStack(self, protocol_template, interlacecheck_yml):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "INTERLACE_Check":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)
    
    ## Options:
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)    
    # overwrite
    self.overwrite = QCheckBox("Overwrite")
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    self.writeimage.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.writeimage, 0, 2)
    
    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QFormLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # correlationThresholdBaseline
    self.correlationThresholdBaseline = QDoubleSpinBox()
    self.correlationThresholdBaseline.setAlignment(Qt.AlignRight)
    self.correlationThresholdBaseline.setSingleStep(0.01)
    self.correlationThresholdBaseline.setDecimals(3)
    self.correlationThresholdBaseline.setStatusTip(interlacecheck_yml["protocol"]["correlationThresholdBaseline"]["description"])
    protocol_layout.addRow(interlacecheck_yml["protocol"]["correlationThresholdBaseline"]["caption"], self.correlationThresholdBaseline) 
    self.correlationThresholdBaseline.valueChanged.connect(self.GetParams)
    # correlationThresholdGradient
    self.correlationThresholdGradient = QDoubleSpinBox()
    self.correlationThresholdGradient.setAlignment(Qt.AlignRight)
    self.correlationThresholdGradient.setSingleStep(0.001)
    self.correlationThresholdGradient.setDecimals(3)
    self.correlationThresholdGradient.setStatusTip(interlacecheck_yml["protocol"]["correlationThresholdGradient"]["description"])
    protocol_layout.addRow(interlacecheck_yml["protocol"]["correlationThresholdGradient"]["caption"], self.correlationThresholdGradient)   
    self.correlationThresholdGradient.valueChanged.connect(self.GetParams)
    # correlationDeviationBaseline
    self.correlationDeviationBaseline = QDoubleSpinBox()
    self.correlationDeviationBaseline.setAlignment(Qt.AlignRight)
    self.correlationDeviationBaseline.setSingleStep(0.1)
    self.correlationDeviationBaseline.setDecimals(3)
    self.correlationDeviationBaseline.setStatusTip(interlacecheck_yml["protocol"]["correlationDeviationBaseline"]["description"])
    protocol_layout.addRow(interlacecheck_yml["protocol"]["correlationDeviationBaseline"]["caption"], self.correlationDeviationBaseline)   
    self.correlationDeviationBaseline.valueChanged.connect(self.GetParams)
    # correlationDeviationGradient
    self.correlationDeviationGradient = QDoubleSpinBox()
    self.correlationDeviationGradient.setAlignment(Qt.AlignRight)
    self.correlationDeviationGradient.setSingleStep(0.1)
    self.correlationDeviationGradient.setDecimals(3)
    self.correlationDeviationGradient.setStatusTip(interlacecheck_yml["protocol"]["correlationDeviationGradient"]["description"])
    protocol_layout.addRow(interlacecheck_yml["protocol"]["correlationDeviationGradient"]["caption"], self.correlationDeviationGradient)         
    self.correlationDeviationGradient.valueChanged.connect(self.GetParams)
    # translationThreshold
    self.translationThreshold = QDoubleSpinBox()
    self.translationThreshold.setAlignment(Qt.AlignRight)
    self.translationThreshold.setSingleStep(0.1)
    self.translationThreshold.setDecimals(3)
    self.translationThreshold.setStatusTip(interlacecheck_yml["protocol"]["translationThreshold"]["description"])
    protocol_layout.addRow(interlacecheck_yml["protocol"]["translationThreshold"]["caption"], self.translationThreshold)    
    self.translationThreshold.valueChanged.connect(self.GetParams)
    # rotationThreshold
    self.rotationThreshold = QDoubleSpinBox()
    self.rotationThreshold.setAlignment(Qt.AlignRight)
    self.rotationThreshold.setSingleStep(0.1)
    self.rotationThreshold.setDecimals(3)
    self.rotationThreshold.setStatusTip(interlacecheck_yml["protocol"]["rotationThreshold"]["description"])
    protocol_layout.addRow(interlacecheck_yml["protocol"]["rotationThreshold"]["caption"], self.rotationThreshold)    
    self.rotationThreshold.valueChanged.connect(self.GetParams)

    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

  def GetParams(self):
    params = [
      'Interlace Correlation Check',
      'INTERLACE_Check', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        'write_image': self.writeimage.isChecked()
        }, 
      'protocol': {
        'correlationDeviationBaseline': self.correlationDeviationBaseline.value(), 
        'correlationDeviationGradient': self.correlationDeviationGradient.value(), 
        'correlationThresholdBaseline': self.correlationThresholdBaseline.value(), 
        'correlationThresholdGradient': self.correlationThresholdGradient.value(), 
        'rotationThreshold': self.rotationThreshold.value(), 
        'translationThreshold': self.translationThreshold.value()
        }
      }
    ]
    self.communicate.SendParams(params)



#
#   utils.py 
#   2021-05-10
#   Written by SK Park, NIRAL, UNC
#
#   Atlasbuilding utilities
#


import os # To run a shell command : os.system("[shell command]")
import sys # to return an exit code
import shutil # to remove a non empty directory
import json
import argparse
import csv 
import copy 
import xml.etree.cElementTree as ET 
from pathlib import Path 

import dtiplayground.dmri.atlasbuilder as ab 
import dtiplayground.dmri.common.tools as tools

logger=ab.logger.write

def generateGreedyAtlasParametersFile(cfg):
    xmlfile=cfg["m_GreedyAtlasParametersTemplatePath"]
    x=ET.parse(xmlfile)
    r=x.getroot()

    ## remove all dummy dataset files
    wis=r.find('WeightedImageSet')
    wi_list=wis.findall('WeightedImage')
    for w in wi_list:
        wis.remove(w)

    ## insert new dataset
    for cid in cfg["m_CasesIDs"]:
        wi=ET.Element('WeightedImage',{})
        lastLoop=str(cfg['m_nbLoops'])
        p=os.path.join(cfg['m_OutputPath'],"1_Affine_Registration/Loop"+lastLoop+"/"+cid+"_Loop"+lastLoop+"_Final"+cfg["m_ScalarMeasurement"]+".nrrd")
        wiFilename=ET.Element('Filename',{'val':str(p)})
        wiItkTransform=ET.Element('ItkTransform',{'val':'1'})
        wi.insert(0,wiFilename)
        wi.insert(1,wiItkTransform)
        wis.insert(-1,wi)  ## insert to the last

    ## change output path 
    for neighbor in r.iter('OutputPrefix'):
        logger("{} {}".format(neighbor.tag,neighbor.attrib))
        neighbor.set('val',cfg["m_OutputPath"]+"/2_NonLinear_Registration/")

    outputfile=cfg["m_OutputPath"]+"/2_NonLinear_Registration/GreedyAtlasParameters.xml"
    x.write(outputfile)


def DisplayErrorAndQuit ( Error ):
    msg='\n\nERROR DETECTED IN WORKFLOW:'+Error
    logger(msg)
    logger('ABORT')
    raise Exception(msg)


# Function that checks if file exist and replace old names by new names if needed
def CheckFileExists ( File, case, caseID ) : # returns 1 if file exists or has been renamed and 0 if not
    if os.path.isfile( File ) : # file exists
      return 1
    else : # file does not exist: check if older version of file can exist (if file name has been changed)
      NamesThatHaveChanged = ["MeanImage", "DiffeomorphicDTI", "DiffeomorphicAtlasDTI", "HField", "GlobalDisplacementField"] # latest versions of the names
      if any( changedname in File for changedname in NamesThatHaveChanged ) : # if name has been changed, check if older version files exist
        if "MeanImage" in File :
          OldFile = File.replace("Mean", "Average")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            return 1
          else:
            return 0
        if "DiffeomorphicDTI" in File :
          OldFile = File.replace( caseID, "Case" + str(case+1) ).replace("DiffeomorphicDTI", "AWDTI")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            os.rename(OldFile.replace("AWDTI","AW"+config['m_ScalarMeasurement']), File.replace("DiffeomorphicDTI","Diffeomorphic"+config['m_ScalarMeasurement']))
            os.rename(OldFile.replace("AWDTI","AWDTI_float"), File.replace("DiffeomorphicDTI","DiffeomorphicDTI_float"))
            return 1
          else : # test other old name
            OldFile = File.replace( caseID, "Case" + str(case+1) )
            if os.path.isfile( OldFile ) : # old file exists: rename and return 1
              os.rename(OldFile, File)
              os.rename(OldFile.replace("DiffeomorphicDTI","Diffeomorphic"+config['m_ScalarMeasurement']), File.replace("DiffeomorphicDTI","Diffeomorphic"+config['m_ScalarMeasurement']))
              os.rename(OldFile.replace("DiffeomorphicDTI","DiffeomorphicDTI_float"), File.replace("DiffeomorphicDTI","DiffeomorphicDTI_float"))
              return 1
            else:
              return 0
        if "DiffeomorphicAtlasDTI" in File :
          OldFile = File.replace("DiffeomorphicAtlasDTI", "AWAtlasDTI")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            os.rename(OldFile.replace("AWAtlasDTI","AWAtlas"+config['m_ScalarMeasurement']), File.replace("DiffeomorphicAtlasDTI","DiffeomorphicAtlas"+config['m_ScalarMeasurement']))
            os.rename(OldFile.replace("AWAtlasDTI","AWAtlasDTI_float"), File.replace("DiffeomorphicAtlasDTI","DiffeomorphicAtlasDTI_float"))
            return 1
          else:
            return 0
        if "HField" in File :
          OldFile = File.replace( caseID, "Case" + str(case+1) ).replace("H", "Deformation")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            return 1
          else : # test other old name
            OldFile = File.replace( caseID, "Case" + str(case+1) )
            if os.path.isfile( OldFile ) : # old file exists: rename and return 1
              os.rename(OldFile, File)
              return 1
            else:
              return 0
        if "GlobalDisplacementField" in File :
          OldFile = File.replace( caseID, "Case" + str(case+1) ).replace("Displacement", "Deformation")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            return 1
          else : # test other old name
            OldFile = File.replace( caseID, "Case" + str(case+1) )
            if os.path.isfile( OldFile ) : # old file exists: rename and return 1
              os.rename(OldFile, File)
              return 1
            else:
              return 0
      else: # file does not exist and name has not been changed: check if the caseX version exists
        if caseID : # CaseID is empty for averages
          OldFile = File.replace( caseID, "Case" + str(case+1) )
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            return 1
          else:
            return 0
        else: # for averages
          return 0

def _check_deformation_sequence_file(config,payload):
    binaryPath=config["m_SoftPath"][9]
    if(not os.path.exists(binaryPath)):
        logger("Software is missing : %s" %binaryPath)
        return False
    logger("Software path is : %s " % binaryPath)
    res=True
    for s in payload:
        cnt=0
        for f in s["filelist"]:
            if(not os.path.exists(f)):
                logger("Error while checking : %s " % f)
                cnt+=1
                res=False
        if cnt==0:
            logger("%s : OK" % s["id"])
        else:
            logger("%s : NOT OK" %s['id'])
    return res 

def _generate_concatenated_displacement_directory(config):
    projectDir=config["m_OutputPath"]
    outDir=os.path.join(projectDir,"displacement_fields")
    if not os.path.isdir(outDir):
      logger("\n=> Creation of the directory containing concatenated displacement fields = " + outDir)
      os.mkdir(outDir)
    return outDir

def ITKTransformTools_Concatenate(config,payload): ## payload should be deformation_track.json list
    command=""
    if(_check_deformation_sequence_file(config,payload)):
        outputDir=_generate_concatenated_displacement_directory(config)
        binaryPath=config["m_SoftPath"][9] 
        command+=binaryPath + " concatenate "
        refDTI=config["m_OutputPath"] + "/final_atlas/5_Final_Atlas/FinalAtlasDTI.nrrd"
        for idx, elm in enumerate(payload):
            hpairList=elm["id"].split("/")
            outFilename="_".join(hpairList) + "_GlobalDisplacementField_Concatenated.nrrd"
            outFilename=os.path.join(outputDir,outFilename)
            if Path(outFilename).exists() and (not config['m_Overwrite']==1):
                logger("File : {} already exists.".format(outFilename))
                continue
            #logger("Output filename : %s"%outFilename)
            tmpCommand=command + outFilename +" -r " + refDTI + " "
            inpListStr=""
            fl=map(str,elm["filelist"])
            #fl.reverse()
            for fn in fl:
                inpListStr+= fn + " displacement "
            tmpCommand+=inpListStr
            #logger("%d : %s " %(idx,tmpCommand))
            os.system(tmpCommand)
    else:
        logger("There are some missing deformation fields file(s)")
        raise(Exception("There are some missing deformation fields file(s)"))

def ITKTransformTools_Concatenate_Inverse(config,payload): ## payload should be deformation_track.json list
    command=""
    if(_check_deformation_sequence_file(config,payload)):
        outputDir=_generate_concatenated_displacement_directory(config)
        binaryPath=config["m_SoftPath"][9] 
        command+=binaryPath + " concatenate "
        for idx, elm in enumerate(payload):
            refDTI=elm['original_dti_path']
            hpairList=elm["id"].split("/")
            outFilename=elm['output_path']
            if Path(outFilename).exists() and (not config['m_Overwrite']==1):
                logger("File : {} already exists.".format(outFilename))
                continue
            #logger("Output filename : %s"%outFilename)
            tmpCommand=command + outFilename +" -r " + refDTI + " "
            inpListStr=""
            fl=map(str,elm["filelist"])
            #fl.reverse()
            for fn in fl:
                inpListStr+= fn + " displacement "
            tmpCommand+=inpListStr
            #logger("%d : %s " %(idx,tmpCommand))
            os.system(tmpCommand)
    else:
        logger("There are some missing deformation fields file(s)")
        raise(Exception("There are some missing deformation fields file(s)"))



def unique(list1): 
    unique_list = [] 
    for x in list1: 
        # check if exists in unique_list or not 
        if x not in unique_list: 
            unique_list.append(x) 
    return unique_list

def isComponent(seq,name):
    comp=list(filter(lambda x : x['name']==name,seq))
    if len(comp)>0 :
        return comp[0] 
    else:
        return False 

def find_config_by_nodename(build_sequence,nodename):
    for cfg in build_sequence:
        if cfg["m_NodeName"]==nodename:
            return cfg 


def generate_deformation_track(seq,node="target"): #input : initialSequence to generate deformation field tracking information (to concatenate them)
    component=isComponent(seq,node)
    outseq=[]

    if component != False:
        for c in component["dataset_ids"]:
            tmpseq=generate_deformation_track(seq,c)
            for t in tmpseq:
                outseq.append(node+"/"+t)
    else:
        outseq.append(node)
        return outseq 
    return outseq

def invert_deformation_track(deformation_seq):
    seq=copy.deepcopy(deformation_seq)
    outseq=[]
    for s in seq:
        elm=s
        strvec=s['id'].split("/")
        strvec.reverse()
        elm['id']='/'.join(strvec)
        #elm['original_dti_id']=strvec[-1]
        arr=[]
        for e in s['filelist']:
            basedir=os.path.dirname(e)
            name="_".join(os.path.basename(e).split('_')[:-1])+"_InverseGlobalDisplacementField.nrrd"
            inverted_deform_path=os.path.join(basedir,name)
            arr.append(inverted_deform_path)
        arr.reverse()
        elm['filelist']=arr 

        output_dir=os.path.dirname(s['output_path'])
        output_name="_".join(os.path.basename(s['output_path']).split('_')[:-2])+"_InverseGlobalDisplacementField_Concatenated.nrrd"
        output_path=os.path.join(output_dir,output_name)
        elm['output_path']=output_path
        outseq.append(elm)
    return outseq 

    
def furnish_deformation_track(seq,project_path,build_sequence,inverse=False): #input deformSequence 
    res=[]
    for d in seq:
        tmp={}
        tmp['id']=d
        compseq=d.split('/')
        cfg=find_config_by_nodename(build_sequence,compseq[-2])
        originalDTIId=compseq[-1]
        originalDTIPath=None
        for idx,case in enumerate(zip(cfg["m_CasesIDs"],cfg["m_CasesPath"])):
            caseID,casePath=case 
            if originalDTIId==caseID: 
                originalDTIPath=casePath
                break

        entry=[]
        for idx,c in enumerate(compseq[0:-1]):
            fpath="atlases/" + c + "/5_Final_Atlas/FinalDeformationFields/" + compseq[idx+1] + "_GlobalDisplacementField.nrrd"
            fpath=os.path.join(project_path,fpath)
            entry.append(fpath)
        tmp['filelist']=entry
        tmp['original_dti_path']=originalDTIPath 
        tmp['original_dti_id']=originalDTIId
        tmp['scalar_measurement']=cfg["m_ScalarMeasurement"]
        tmp['nb_loops']=cfg['m_nbLoops']
        tmp['nb_loops_dtireg']=cfg['m_nbLoopsDTIReg']
        tmp['project_path']=cfg['m_OutputPath']
        tmp['need_to_be_cropped']=cfg['m_NeedToBeCropped']
        outputDir=os.path.join(project_path,"displacement_fields")
        hpairList=tmp["id"].split("/")
        outFilename="_".join(hpairList) + "_GlobalDisplacementField_Concatenated.nrrd"
        outFilename=os.path.join(outputDir,outFilename)
        tmp['output_path']=outFilename
        res.append(tmp)
    return res 




def parse_hbuild(hb,root_path,root_node="target"): #hbuild parser to generate build sequence
    if root_node is None:
        root_node=hb['project']['target_node']
    root=hb['build'][root_node]
    seq=[]
    nodeFiles=[] ## sub node's final atlases
    # scalar=hb['config']['m_ScalarMeasurement']
    if root["type"]=="node":    
        for c in root["components"]:
            seq+=parse_hbuild(hb, root_path=root_path, root_node=c)
            nodeAtlasPath=os.path.join(root_path,"atlases/"+c+"/5_Final_Atlas/FinalAtlasDTI.nrrd")
            nodeFiles.append(nodeAtlasPath)
    elif root["type"]=="end_node":
        if root["filetype"]=="dataset":
            rows=[]
            rows_id=[]
            with open(str(root['datasetfiles']),'r') as f:
                csvreader=csv.reader(f)
                next(csvreader,None)
                for r in csvreader:
                    fpath=str(r[1])
                    fid=os.path.splitext(os.path.basename(fpath))[0]
                    rows.append(fpath)
                    rows_id.append(str(fid))

            return  [{"name" : str(root_node),
                "dataset_files" : rows,
                "dataset_ids" : rows_id,
                "project_path" : str(os.path.join(root_path,"atlases/"+root_node))
                }]
        else:
            flist=list(map(str,root["datasetfiles"]))
            fids=[]
            for e in flist:
                fid=os.path.splitext(os.path.basename(e))[0]
                fids.append(fid)

            return [{"name" : str(root_node),
                    "dataset_files" : flist,
                    "dataset_ids" : fids ,
                    "project_path" : str(os.path.join(root_path,"atlases/"+root_node))
                    }]

    # node type file reading

    seq+=[{"name" : str(root_node),
            "dataset_files" : list(map(str,nodeFiles)),
            "dataset_ids" : list(map(str,root["components"])),
            "project_path" : str(os.path.join(root_path,"atlases/"+root_node))

         }]
    seq=unique(seq)

    ## generate final buildsequence furnished with configuration


    return seq

def furnish_sequence(hb,seq):
    bs=[]
    for s in seq:
        conf=hb["config"].copy()
        conf["m_OutputPath"]=s['project_path']
        conf["m_CasesPath"]=s['dataset_files']
        conf["m_CasesIDs"]=s['dataset_ids']
        conf["m_NodeInfo"]=hb["build"][s['name']]
        conf["m_NodeName"]=s["name"]
        bs.append(conf)

    return bs

def generate_directories(project_path,sequence): ## from build sequence, generate directories
    atlasesPath=os.path.join(project_path,"atlases")
    finalAtlasPath=os.path.join(project_path,"final_atlas")
    if not os.path.isdir(atlasesPath):
      logger("\n=> Creation of the atlas directory = " + atlasesPath)
      os.mkdir(atlasesPath)
    if not os.path.isdir(finalAtlasPath):
      logger("\n=> Creation of the atlas directory = " + finalAtlasPath)
      os.mkdir(finalAtlasPath)
    for s in sequence:
        apath=os.path.join(s["m_OutputPath"])
        if not os.path.isdir(apath):
          logger("\n=> Creation of the atlas directory = " + apath)
          os.mkdir(apath)
    logger("Initial directories are generated")


def dependency_satisfied(hb,node_name,completed_atlases):
    if hb["build"][node_name]["type"]=="end_node": 
        return True
    else:
        comps=hb["build"][node_name]["components"]
        for c in comps:
            if c not in completed_atlases: return False 
        return True



def generate_results_csv_from_deformation_track(deformation_track,project_path): # generate final result file with deformation track file

    dt=deformation_track
    outpath=os.path.join(project_path,"DTIAtlasBuilderResults.csv")
    
    m_ScalarMeasurement=dt[0]["scalar_measurement"]
    m_NeedToBeCropped=dt[0]["need_to_be_cropped"]
    header=["id", "Original DTI Image"]
    if m_NeedToBeCropped==1: header + ["Cropped DTI"]
    tmp=[m_ScalarMeasurement+ " from original",
        "Affine transform", "Affine Registered DTI", 
        "Affine Registered "+m_ScalarMeasurement,
        "Diffeomorphic Deformed " + m_ScalarMeasurement,
        "Diffeomorphic Deformation field to Affine space",
        "Diffeomorphic Deformation field to Affine space",
        "Diffeomorphic DTI",
        "Diffeomorphic Deformation field to Original space",
        "DTI-Reg Final DTI"
        ]
    header+=tmp
    with open(outpath,"w") as f:
        csvwriter=csv.writer(f,delimiter=',')
        csvwriter.writerow(header)
        for idx,case in enumerate(dt):
            caseID,casePath = case["original_dti_id"],case["original_dti_path"]
            m_OutputPath=case["project_path"]
            m_nbLoops=case["nb_loops"]
            m_nbLoopsDTIReg=case["nb_loops_dtireg"]
            row=[
                idx+1,
                casePath]
            if m_NeedToBeCropped==1: row+=[m_OutputPath+"/1_Affine_Registration/" + caseID+"_croppedDTI.nrrd"]
            concatenated_displacement_path=case["output_path"]
            row+=[
                m_OutputPath+"/1_Affine_Registration/" + caseID + "_" + m_ScalarMeasurement + ".nrrd",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_LinearTrans.txt",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_LinearTrans_DTI.nrrd",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_Final" + m_ScalarMeasurement +".nrrd",
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_NonLinearTrans_" + m_ScalarMeasurement + ".mhd",
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_HField.mhd" ,
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_InverseHField.mhd" ,
                m_OutputPath+"/3_Diffeomorphic_Atlas/" + caseID + "_DiffeomorphicDTI.nrrd",
                concatenated_displacement_path,
                m_OutputPath+"/4_Final_Resampling/FinalTensors/" + caseID + "_FinalDeformedDTI.nrrd"
            ]
            csvwriter.writerow(row)

def generate_results_csv(cfg):

    outpath=os.path.join(cfg["m_OutputPath"],"DTIAtlasBuilderResults.csv")
    m_OutputPath=cfg["m_OutputPath"]
    m_ScalarMeasurement=cfg["m_ScalarMeasurement"]
    m_nbLoops=cfg["m_nbLoops"]
    m_nbLoopsDTIReg=cfg["m_nbLoopsDTIReg"]
    m_NeedToBeCropped=cfg["m_NeedToBeCropped"]
    header=["id", "Original DTI Image"]
    if m_NeedToBeCropped==1: header + ["Cropped DTI"]
    tmp=[cfg["m_ScalarMeasurement"]+ " from original",
        "Affine transform", "Affine Registered DTI", 
        "Affine Registered "+cfg["m_ScalarMeasurement"],
        "Diffeomorphic Deformed " + cfg["m_ScalarMeasurement"],
        "Diffeomorphic Deformation field to Affine space",
        "Diffeomorphic Deformation field to Affine space",
        "Diffeomorphic DTI",
        "Diffeomorphic Deformation field to Original space",
        "DTI-Reg Final DTI"
        ]
    header+=tmp
    with open(outpath,"w") as f:
        csvwriter=csv.writer(f,delimiter=',')
        csvwriter.writerow(header)
        for idx,case in enumerate(zip(cfg["m_CasesIDs"],cfg["m_CasesPath"])):
            caseID,casePath = case
            row=[
                idx+1,
                casePath]
            if m_NeedToBeCropped==1: row+=[m_OutputPath+"/1_Affine_Registration/" + caseID+"_croppedDTI.nrrd"]
            row+=[
                m_OutputPath+"/1_Affine_Registration/" + caseID + "_" + m_ScalarMeasurement + ".nrrd",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_LinearTrans.txt",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_LinearTrans_DTI.nrrd",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_Final" + m_ScalarMeasurement +".nrrd",
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_NonLinearTrans_" + m_ScalarMeasurement + ".mhd",
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_HField.mhd" ,
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_InverseHField.mhd" ,
                m_OutputPath+"/3_Diffeomorphic_Atlas/" + caseID + "_DiffeomorphicDTI.nrrd",
                m_OutputPath+"/4_Final_Resampling/FinalDeformationFields/" + caseID + "_GlobalDisplacementField.nrrd",
                m_OutputPath+"/4_Final_Resampling/FinalTensors/" + caseID + "_FinalDeformedDTI.nrrd"
            ]
            csvwriter.writerow(row)




